
SongbookWindow = class( Turbine.UI.Window );
selectedSong = ""; -- set the default song
selectedSongIndex = 1;
selectedTrack = 1;
selectedDir = "/"; -- set the default dir
dirPath = {}; -- table holding directory path
dirPath[1] = "/"; -- set first item as root dir
librarySize = 0;

ListBoxScrolled = class( Turbine.UI.ListBox ); -- Listbox with child scrollbar and separator
ListBoxCharColumn = class( ListBoxScrolled ) -- Listbox with single char column

-- fix to prevent Vindar patch from messing up anything since it's not needed
SongbookLoad = Turbine.PluginData.Load;
SongbookSave = Turbine.PluginData.Save;

Settings = { WindowPosition = { Left = "700", Top = "20", Width = "342", Height = "398" }, WindowVisible = "no", WindowOpacity="0.9", DirHeight = "100", TracksHeight = "50", TracksVisible = "no", ToggleVisible = "yes", ToggleLeft = tostring(Turbine.UI.Display.GetWidth()-55), ToggleTop = "310", ToggleOpacity = "0.25", SearchVisible = "yes", DescriptionVisible = "no", DescriptionFirst = "no" }; -- default values
CharSettings = {
};

-- if (lang == "de" or lang == "fr") then	
	-- if (Turbine.Engine.GetLocale() == "de" or Turbine.Engine.GetLocale() == "fr") then
		-- Settings.WindowOpacity = "0,9";
	-- end
-- end
euroFormat=(tonumber("1,000")==1);
if euroFormat then
	Settings.WindowOpacity = "0,9";
	Settings.ToggleOpacity = "0,25";
end

SongDB = {
	Directories = {
	},
	Songs = {	
	}
};

function SongbookWindow:Constructor()
	Turbine.UI.Lotro.Window.Constructor( self );
	
	SongDB = SongbookLoad( Turbine.DataScope.Account , "SongbookData") or SongDB;
	Settings = SongbookLoad( Turbine.DataScope.Account , gSettings) or Settings;
	CharSettings = SongbookLoad( Turbine.DataScope.Character , gSettings) or CharSettings;

	-- Variables for filters and setups
	self.sFilterPartcount = nil -- A char for every acceptable part count, with 'A' being solo, 'B' two parts, etc.
	self.maxTrackCount = 25 -- Assumed maximum number of track setups (adjust if necessary):

	self.bFilter = false -- show/hide filter UI
	self.bChiefMode = true -- enables sync start shortcut, uses party object (seems to work for FS leader)
	self.bShowPlayers = true -- show/hide players listbox (used to auto-hide, but disabled for now)
	self.aFilteredIndices = { } -- Array for filtered indices, k = display index; v = SongDB index
	self.aPlayers = { } -- k = player name, v = ready track, 0 if no track ready
	self.nPlayers = 0 -- number of players (unfortunately not as simple as #self.aPlayers)
	self.aCurrentSongReady = { } -- k = player name; v = track ready state (see GetTrackReadyState())
	self.aReadyTracks = "" -- indicates which tracks are ready (A = 1st, B = 2nd, etc). Used for setup checks
	self.aSetupTracksIndices = { } -- when tracks are filtered for a setup, this array contains track indices
	self.aSetupListIndices = { } -- list index for tracks that are part of the currently selected setup
	self.iCurrentSetup = nil -- indicates which setup is currently selected
	self.selectedSetupCount = 'A' -- Stores the code of the current setup to select it on song change, if available
	self.maxPartCount = nil -- the number of parts to use as filter (nil if not filtering, else player count)
	self.alignTracksRight = false -- if true, track names are listed right-aligned (resize will reset to left aligned)
	self.listboxSetupsWidth = 20 -- width of the setups listbox (to the left of the tracks list)
	self.setupsWidth = self.listboxSetupsWidth + 10 -- total width of setup list including scrollbar
	self.bShowSetups = false -- show/hide setups (autohide for songs with no setups defined)

	-- colours for the different track/player ready states in the track and player listboxes
	self.colourDefaultHighlighted = Turbine.UI.Color( 1, 0.15, 0.95, 0.15 )
	self.colourReadyHighlighted = Turbine.UI.Color( 1, 0.15, 0.60, 0.15 )
	self.colourReadyMultipleHighlighted = Turbine.UI.Color( 1, 0.7, 0.7, 1 )
	self.colourDefault = Turbine.UI.Color( 1, 1, 1, 1 )
	self.colourReady = Turbine.UI.Color( 1, 0.4, 0.4, 0 )
	self.colourReadyMultiple = Turbine.UI.Color( 1, 0.6, 0.6, 0.95 )
	self.colourDifferentSong = Turbine.UI.Color( 1, 0, 0 )
	self.colourDifferentSetup = Turbine.UI.Color( 1, 0.6, 0 )
	self.colourWrongInstrument = Turbine.UI.Color( 1, 0.6, 0 )
	self.backColourDefault = Turbine.UI.Color( 1, 0, 0, 0 )
	self.backColourHighlight = Turbine.UI.Color( 1, 0.1, 0.1, 0.1 )
	--self.backColourWrongInstrument = Turbine.UI.Color( 1, 0.25, 0.1, 0.1 )
	
	self.bCheckInstrument = true
	self.bInstrumentOk = true
	self.aInstruments = { "bagpipe", "clarinet", "cowbell", "drum", "flute", "harp", "horn", "lute", "pibgorn", "theorbo"  }
	self.aSpecialInstruments = { }
	self.aSpecialInstruments[ "satakieli" ] = 6 -- index in the insturments array
	self.bTimer = false
	self.bTimerCountdown = false
	
	self.bShowReadyChars = true
	self.bHighlightReadyCol = false
	self.chNone = " "
	self.chReady = "~"
	self.chWrongSong = "S"
	self.chWrongPart = "P"
	self.chMultiple = "M"
	
	-- Legacy fixes
	if not Settings.DirHeight then
		Settings.DirHeight = "100";
	end
	if not Settings.TracksHeight then
		Settings.TracksHeight = "40";
	end
	if not Settings.TracksVisible then
		Settings.TracksVisible = "no";
	end
	if not WindowVisible then
		WindowVisible = "no";
	end	
	if not Settings.SearchVisible then
		Settings.SearchVisible = "yes";	
	end
	if not Settings.DescriptionVisible then
		Settings.DescriptionVisible = "no";	
	end
	if not Settings.DescriptionFirst then
		Settings.DescriptionFirst = "no";	
	end
	
	if not Settings.ToggleOpacity then
		Settings.ToggleOpacity = 1/4;
	end		

	if not Settings.FiltersState then Settings.FiltersState = "false"; end		
	if not Settings.ChiefMode then Settings.ChiefMode = "true"; end		
	if not Settings.TimerState then Settings.TimerState = "false"; end
	if not Settings.TimerCountdown then Settings.TimerCountdown = "false"; end		
 	if not Settings.ReadyColState then Settings.ReadyColState = "false"; end
 	if not Settings.ReadyColHighlight then Settings.ReadyColHighlight = "false"; end
	
	if not SongDB.Songs then
		SongDB = {
			Directories = {
			},
			Songs = {	
			}
		};
	end		
	
	if not Settings.Commands then
		Settings.Commands = {};		
		Settings.Commands["1"] = { Title = Strings["cmd_demo1_title"], Command = Strings["cmd_demo1_cmd"] };
		Settings.Commands["2"] = { Title = Strings["cmd_demo2_title"], Command = Strings["cmd_demo2_cmd"] };
		Settings.Commands["3"] = { Title = Strings["cmd_demo3_title"], Command = Strings["cmd_demo3_cmd"] };
		Settings.DefaultCommand = "1";
	end
	
	if not CharSettings.InstrSlots then
		CharSettings.InstrSlots = {};
		CharSettings.InstrSlots["visible"] = "yes";
		CharSettings.InstrSlots["number"] = 8;
		for i = 1, CharSettings.InstrSlots["number"] do
			CharSettings.InstrSlots[tostring(i)] = { qsType = "", qsData = "" };		
		end
	end
	if not CharSettings.InstrSlots["number"] then
		CharSettings.InstrSlots["number"] = 8;
	end
	for i = 1, CharSettings.InstrSlots["number"] do
		CharSettings.InstrSlots[tostring(i)].qsType = tonumber(CharSettings.InstrSlots[tostring(i)].qsType);
	end		
	
	-- unstringify settings
	Settings.WindowPosition.Left = tonumber(Settings.WindowPosition.Left);
	Settings.WindowPosition.Top = tonumber(Settings.WindowPosition.Top);
	Settings.WindowPosition.Width = tonumber(Settings.WindowPosition.Width);
	Settings.WindowPosition.Height = tonumber(Settings.WindowPosition.Height);
	self:ValidateWindowPosition( Settings.WindowPosition )
	Settings.ToggleTop = tonumber(Settings.ToggleTop);
	Settings.ToggleLeft = tonumber(Settings.ToggleLeft);
	Settings.DirHeight = tonumber(Settings.DirHeight);
	Settings.TracksHeight = tonumber(Settings.TracksHeight);
	Settings.WindowOpacity = tonumber(Settings.WindowOpacity);
	Settings.ToggleOpacity = tonumber(Settings.ToggleOpacity);
	CharSettings.InstrSlots["number"] = tonumber(CharSettings.InstrSlots["number"]);

	-- Fix to prevent window or toggle to travel outside of the screen
	local displayWidth, displayHeight = Turbine.UI.Display.GetSize();
	if Settings.WindowPosition.Left + Settings.WindowPosition.Width > displayWidth then
		Settings.WindowPosition.Left = displayWidth - Settings.WindowPosition.Width;
	end
	if Settings.WindowPosition.Top + Settings.WindowPosition.Height > displayHeight then
		Settings.WindowPosition.Top = displayHeight - Settings.WindowPosition.Height;
	end
	if Settings.WindowPosition.Left < 0 then
		Settings.WindowPosition.Left = 0;
	end
	if Settings.WindowPosition.Top < 0 then
		Settings.WindowPosition.Top = 0;
	end	
	if Settings.ToggleLeft + 35 > displayWidth then
		Settings.ToggleLeft = displayWidth - 35;
	end
	if Settings.ToggleTop + 35 > displayHeight then
		Settings.ToggleTop = displayHeight - 35;
	end
	if Settings.ToggleLeft < 0 then
		Settings.ToggleLeft = 0;
	end
	if Settings.ToggleTop < 0 then
		Settings.ToggleTop = 0;
	end
	
	
	-- Hide UI when F12 is pressed
	local hideUI = false;
	local wasVisible;
	self:SetWantsKeyEvents(true);
	self.KeyDown = function(sender, args)
		if (args.Action == 268435635) then
			if not hideUI then
				hideUI = true;
				if self:IsVisible() then
					wasVisible = true;
					self:SetVisible(false);					
				else
					wasVisible = false;
				end
				settingsWindow:SetVisible(false);
				toggleWindow:SetVisible(false);
			else
				hideUI = false;
				if wasVisible then
					self:SetVisible(true);
					settingsWindow:SetVisible(false);					
				end
				if (Settings.ToggleVisible == "yes") then
					toggleWindow:SetVisible(true);
				end
			end
		end
	end
	
	librarySize = #SongDB.Songs;
	
	self:SetPosition( Settings.WindowPosition.Left, Settings.WindowPosition.Top );
	self:SetSize( Settings.WindowPosition.Width, Settings.WindowPosition.Height );
	--self:SetZOrder(10);
	
	self:SetOpacity( Settings.WindowOpacity );
	self:SetText("Songbook " .. Plugins[gPlugin]:GetVersion() .. Strings["title"] );
	
	self.minWidth = 342;
	self.minHeight = 308;
	self.lFXmod = 23; -- listFrame x coord modifier
	self.lCXmod = 28; -- listContainer x coord modifier (original value was 42)

	if (CharSettings.InstrSlots["visible"] == "yes") then
		self.lFYmod = 214; -- listFrame y coord modifier = difference between bottom pixels and window bottom
		self.lCYmod = 233; -- listContainer y coord modifier = difference between bottom pixels and window bottom
	else
		self.lFYmod = 169; -- listFrame y coord modifier = difference between bottom pixels and window bottom
		self.lCYmod = 188; -- listContainer y coord modifier = difference between bottom pixels and window bottom
	end
	
	-- Frame for the song list
	self.listFrame = Turbine.UI.Control();
	self.listFrame:SetParent( self );
	self.listFrame:SetBackColor( Turbine.UI.Color(1, 0.15, 0.15, 0.15) );
	self.listFrame:SetPosition(12, 134);
	--self.listFrame:SetSize(self:GetWidth() - self.lFXmod, self:GetHeight() - self.lFYmod);
	self.listContainer = Turbine.UI.Control();
	self.listContainer:SetParent( self );
	self.listContainer:SetBackColor( Turbine.UI.Color(1,0,0,0) );
	self.listContainer:SetPosition(18, 147);
	--self.listContainer:SetSize(self:GetWidth() - self.lCXmod, self:GetHeight() - self.lCYmod);

	-- outer frame title
	self.listFrame.heading = Turbine.UI.Label();
	self.listFrame.heading:SetParent( self.listFrame );
	self.listFrame.heading:SetLeft(5);
	self.listFrame.heading:SetSize(self.listFrame:GetWidth(),13);
	self.listFrame.heading:SetFont( Turbine.UI.Lotro.Font.TrajanPro13 );
	self.listFrame.heading:SetText( Strings["ui_dirs"] );
	
	-- separator1 between dir list and song list
	self.separator1 = self:CreateMainSeparator( Settings.DirHeight );
	self.separator1:SetVisible(true);
	-- separator1 title
	self.separator1.heading = self:CreateSeparatorHeading( self.separator1, Strings["ui_songs"] );
	-- separator1 hint arrows
	self.sArrows1 = self:CreateSeparatorArrows( self.separator1 )
	
	-- separator2 between song list and track list
	self.sepSongsTracks = self:CreateMainSeparator( self.listContainer:GetHeight() - Settings.TracksHeight - 13 );
	-- separator2 title
	self.sepSongsTracks.heading = self:CreateSeparatorHeading( self.sepSongsTracks, Strings["ui_parts"] );
	-- separator2 hint arrows
	self.sArrows2 = self:CreateSeparatorArrows( self.sepSongsTracks )

	-- Tooltip
	self.tipLabel = Turbine.UI.Label();
	self.tipLabel:SetParent( self );
	self.tipLabel:SetPosition( self:GetWidth() - 270, 27 );
	self.tipLabel:SetSize(245, 30);
	self.tipLabel:SetTextAlignment( Turbine.UI.ContentAlignment.MiddleRight );
	self.tipLabel:SetText("");
	
	-- Music mode button
	self.musicSlot = self:CreateMainShortcut(20);
	-- Trying to fix the problem with unresponsive buttons. Haven't found out yet how to disable
	-- dragging from a quickslot altogether, so for now this just restores the shortcut.
	self.musicSlotShortcut = Turbine.UI.Lotro.Shortcut( Turbine.UI.Lotro.ShortcutType.Alias, Strings["cmd_music"]);
	self.musicSlot.DragDrop =
	function( sender, args )
		if( self.musicSlotShortcut ) then
			self.musicSlot:SetShortcut( self.musicSlotShortcut ); -- restore shortcut
		end
	end
	self.musicSlot:SetShortcut( self.musicSlotShortcut );
	self.musicSlot:SetVisible( true );
	
	-- Play button
	self.playSlot = self:CreateMainShortcut(60);
	self.playSlot.DragDrop =
	function( sender, args )
		if( self.playSlotShortcut ) then
			self.playSlot:SetShortcut( self.playSlotShortcut ); -- restore shortcut
		end
	end
	
	-- Ready check button
	self.readySlot = self:CreateMainShortcut(120);
	self.readySlot:SetShortcut( Turbine.UI.Lotro.Shortcut( Turbine.UI.Lotro.ShortcutType.Alias, Strings["cmd_ready"]));
	
	-- Sync play button
	self.syncSlot = self:CreateMainShortcut(161);
	self.syncSlot.DragDrop =
	function( sender, args )
		if( self.syncSlotShortcut ) then
			self.syncSlot:SetShortcut( self.syncSlotShortcut ); -- restore shortcut
		end
	end
	
	-- Start sync play button
	self.syncStartSlot = self:CreateMainShortcut(202);
	self.syncStartSlotShortcut = Turbine.UI.Lotro.Shortcut( Turbine.UI.Lotro.ShortcutType.Alias, Strings["cmd_start"] );
	self.syncStartSlot.DragDrop =
	function( sender, args )
		if( self.syncStartSlotShortcut ) then
			self.syncStartSlot:SetShortcut( self.syncStartSlotShortcut ); -- restore shortcut
		end
	end
	self.syncStartSlot:SetShortcut(self.syncStartSlotShortcut);
	
	-- Share button
	self.shareSlot = self:CreateMainShortcut(287);
	if (Settings.Commands[Settings.DefaultCommand]) then
		self.shareSlot:SetShortcut( Turbine.UI.Lotro.Shortcut( Turbine.UI.Lotro.ShortcutType.Alias, self:ExpandCmd(Settings.DefaultCommand)));
	end
	
	-- Track label
	self.trackLabel = Turbine.UI.Label();
	self.trackLabel:SetParent( self );
	self.trackLabel:SetPosition(247, 63);
	self.trackLabel:SetSize(30, 12);
	self.trackLabel:SetZOrder(200);
	self.trackLabel:SetText("X:");

	-- Track number
	self.trackNumber = Turbine.UI.Label();
	self.trackNumber:SetParent( self );
	self.trackNumber:SetPosition(262, 63);
	self.trackNumber:SetWidth(20);
	
	-- Track up arrow
	self.trackPrev = Turbine.UI.Control();
	self.trackPrev:SetParent( self );
	self.trackPrev:SetPosition(252, 51);
	self.trackPrev:SetSize(12, 8);
	self.trackPrev:SetBackground(gDir .. "arrowup.tga");
	self.trackPrev:SetBlendMode( Turbine.UI.BlendMode.AlphaBlend );
	self.trackPrev:SetVisible( false );
	
	-- Track down arrow
	self.trackNext = Turbine.UI.Control();
	self.trackNext:SetParent( self );
	self.trackNext:SetPosition(252, 78);
	self.trackNext:SetSize(12, 8);
	self.trackNext:SetBackground(gDir .. "arrowdown.tga");
	self.trackNext:SetBlendMode( Turbine.UI.BlendMode.AlphaBlend );
	self.trackNext:SetVisible( false );
	
	-- actions for track change
	self.trackPrev.MouseClick = function(sender, args)
		if(args.Button == Turbine.UI.MouseButton.Left) then
			self:SelectTrack(selectedTrack - 1);
		end
	end
	self.trackNext.MouseClick = function(sender, args)
		if(args.Button == Turbine.UI.MouseButton.Left) then
			self:SelectTrack(selectedTrack + 1);
		end
	end
		
	-- actions for button mouse hovers
	self.musicSlot.MouseEnter = function(sender,args)
		self.musicIcon:SetBackground(gDir .. "icn_m_hover.tga");
		self.tipLabel:SetText(Strings["tt_music"]);
	end
	self.musicSlot.MouseLeave = function(sender,args)
		self.musicIcon:SetBackground(gDir .. "icn_m.tga");
		self.tipLabel:SetText("");
	end
	self.playSlot.MouseEnter = function(sender,args)
		self.playIcon:SetBackground(gDir .. "icn_p_hover.tga");
		self.tipLabel:SetText(Strings["tt_play"]);
	end
	self.playSlot.MouseLeave = function(sender,args)
		self.playIcon:SetBackground(gDir .. "icn_p.tga");
		self.tipLabel:SetText("");
	end
	self.readySlot.MouseEnter = function(sender,args)
		self.readyIcon:SetBackground(gDir .. "icn_r_hover.tga");
		self.tipLabel:SetText(Strings["tt_ready"]);
	end
	self.readySlot.MouseLeave = function(sender,args)
		self.readyIcon:SetBackground(gDir .. "icn_r.tga");
		self.tipLabel:SetText("");
	end
	self.syncSlot.MouseEnter = function(sender,args)
		self.syncIcon:SetBackground(gDir .. "icn_s_hover.tga");
		self.tipLabel:SetText(Strings["tt_sync"]);
		--ici
	end
	self.syncSlot.MouseLeave = function(sender,args)
		self.syncIcon:SetBackground(gDir .. "icn_s.tga");
		self.tipLabel:SetText("");
	end
	self.syncStartSlot.MouseEnter = function(sender,args)
		self.syncStartIcon:SetBackground(gDir .. "icn_ss_hover.tga");
		self.tipLabel:SetText(Strings["tt_start"]);
	end
	self.syncStartSlot.MouseLeave = function(sender,args)
		self.syncStartIcon:SetBackground(gDir .. "icn_ss.tga");
		self.tipLabel:SetText("");
	end
	self.shareSlot.MouseEnter = function(sender,args)
		self.shareIcon:SetBackground(gDir .. "icn_sh_hover.tga");
		if (Settings.Commands[Settings.DefaultCommand].Title) then
			self.tipLabel:SetText(Settings.Commands[Settings.DefaultCommand].Title);
		end
	end
	self.shareSlot.MouseLeave = function(sender,args)
		self.shareIcon:SetBackground(gDir .. "icn_sh.tga");
		self.tipLabel:SetText("");
	end
	self.shareSlot.MouseWheel = function(sender,args)
		local nextCmd = tonumber(Settings.DefaultCommand) - args.Direction;
		local size = SettingsWindow:CountCmds();
		
		if (nextCmd == 0) then
			Settings.DefaultCommand = tostring(size);		
		elseif (nextCmd > size) then
			Settings.DefaultCommand = "1";		
		else
			Settings.DefaultCommand = tostring(nextCmd);		
		end
		
		self.shareSlot:SetShortcut( Turbine.UI.Lotro.Shortcut( Turbine.UI.Lotro.ShortcutType.Alias, self:ExpandCmd(Settings.DefaultCommand)));		
		self.shareSlot:SetVisible(true);
	end
	self.trackLabel.MouseClick = function(sender,args)
		if(args.Button == Turbine.UI.MouseButton.Left) then
			self:ToggleTracks();
		end
	end
	
	-- icons that hide default quick slots
	self.musicIcon = self:CreateMainIcon(20,"icn_m");
	self.playIcon = self:CreateMainIcon(60,"icn_p");
	self.readyIcon = self:CreateMainIcon(120,"icn_r");
	self.syncIcon = self:CreateMainIcon(161,"icn_s");
	self.syncStartIcon = self:CreateMainIcon(202,"icn_ss");
	self.shareIcon = self:CreateMainIcon(287,"icn_sh");
	
	-- selected song display
	self.songTitle = Turbine.UI.Label();
	self.songTitle:SetParent( self );
	self.songTitle:SetFont(Turbine.UI.Lotro.Font.Verdana16);
	self.songTitle:SetForeColor( self.colourDefaultHighlighted );	
	self.songTitle:SetPosition( 23, 90 );
	self.songTitle:SetSize( self:GetWidth() - 52, 16);
	
	-- search field
	self.searchInput = Turbine.UI.Lotro.TextBox();
	self.searchInput:SetParent(self);
	self.searchInput:SetPosition(17, 110);
	self.searchInput:SetSize(150, 20);
	self.searchInput:SetFont(Turbine.UI.Lotro.Font.Verdana14);
	self.searchInput:SetMultiline(false);	
	self.searchInput:SetVisible( false );
	local searchFocus = false; 
	self.searchInput.KeyDown = function(sender, args)
		if (args.Action == 162) then
			if (searchFocus) then
				self:SearchSongs();
			end
		end
	end
	self.searchInput.FocusGained = function(sender, args)
		searchFocus = true;
	end
	self.searchInput.FocusLost = function(sender, args)
		searchFocus = false;
	end
	
	-- search button
	self.searchBtn = Turbine.UI.Lotro.Button();
	self.searchBtn:SetParent(self);
	self.searchBtn:SetPosition(172, 110);
	self.searchBtn:SetSize(80, 20);
	self.searchBtn:SetText(Strings["ui_search"]);
	self.searchBtn:SetVisible( false );

	self.searchBtn.MouseClick = function(sender, args)
		self:SearchSongs();
	end
	
	-- clear search button
	self.clearBtn = Turbine.UI.Lotro.Button();
	self.clearBtn:SetParent(self);
	self.clearBtn:SetPosition(255, 110);
	self.clearBtn:SetSize(70, 20);
	self.clearBtn:SetText(Strings["ui_clear"]);
	self.clearBtn:SetVisible( false );
	
	self.clearBtn.MouseClick = function(sender, args)
		self.searchInput:SetText("");
		self.songlistBox:ClearItems();
		self:LoadSongs();
		self:SelectSong(1);
	end
	
	-- hide search components if not toggled
	if (Settings.SearchVisible == "yes") then
		self.searchInput:SetVisible( true );
		self.searchBtn:SetVisible( true );
		self.clearBtn:SetVisible( true );				
	end

	-- directory list box
	self.dirlistBox = ListBoxScrolled:New( 10 );
	self.dirlistBox:SetParent( self.listContainer );
	self.dirlistBox:SetVisible( true );
	
	-- track list box
	self.tracklistBox = ListBoxCharColumn:New( 10, 20 );
	self.tracklistBox:SetParent( self.listContainer );
	self:AdjustTracklistLeft( );
	-- show track components if toggled
	self:ShowTrackListbox( Settings.TracksVisible == "yes" )
	
	-- main song list box
	self.songlistBox = ListBoxScrolled:New( 10 );
	self.songlistBox:SetParent( self.listContainer );
	self.songlistBox:SetVisible( true );
	self.songlistBox:SetVisible( true );
	
	-- instrument slot container
	self.instrContainer = Turbine.UI.Control();
	self.instrContainer:SetParent( self );
	self.instrContainer:SetPosition( 10, self:GetHeight() - 75 );
	if (CharSettings.InstrSlots["visible"] == "yes") then
		self.instrContainer:SetVisible( true );
	else
		self.instrContainer:SetVisible( false );
	end
	self.instrContainer:SetSize( 40*CharSettings.InstrSlots["number"], 38 );
	self.instrContainer:SetZOrder(90);

	-- instrument slots
	self.instrSlot = {};
	
	local instrdrag = false;
	for i=1,CharSettings.InstrSlots["number"] do
		self.instrSlot[i] = Turbine.UI.Lotro.Quickslot();
		self.instrSlot[i]:SetParent( self.instrContainer );
		self.instrSlot[i]:SetPosition(40*(i-1), 0);
		self.instrSlot[i]:SetSize(37, 37);
		self.instrSlot[i]:SetZOrder(100);
		self.instrSlot[i]:SetAllowDrop(true);
		
		if (CharSettings.InstrSlots[tostring(i)].data ~= "") then
			pcall(function() 
				local sc = Turbine.UI.Lotro.Shortcut( CharSettings.InstrSlots[tostring(i)].qsType, CharSettings.InstrSlots[tostring(i)].qsData);
				self.instrSlot[i]:SetShortcut(sc);
			end);
		end
		
		self.instrSlot[i].ShortcutChanged = function( sender, args )
			pcall(function() 
				local sc = sender:GetShortcut();
				CharSettings.InstrSlots[tostring(i)].qsType = tostring(sc:GetType());
				CharSettings.InstrSlots[tostring(i)].qsData = sc:GetData();
			end);
		end
		
		self.instrSlot[i].DragLeave = function( sender, args )
			if (instrdrag) then 
				CharSettings.InstrSlots[tostring(i)].qsType ="";
				CharSettings.InstrSlots[tostring(i)].qsData = "";
				local sc = Turbine.UI.Lotro.Shortcut( "", "");
				self.instrSlot[i]:SetShortcut(sc);
				instrdrag = false;
			end
		end
		
		self.instrSlot[i].MouseDown = function( sender, args )
			if(args.Button == Turbine.UI.MouseButton.Left) then	
				instrdrag = true;
			end
		end
	end
	
	-- adjust to search visibility
	if (Settings.SearchVisible == "no") then 
		self:ToggleSearch("off");
	end
	
	self.bTimer = ( Settings.TimerState == "true" )
	self.bTimerCountdown = ( Settings.TimerCountdown == "true" )
	self.bShowReadyChars = ( Settings.ReadyColState == "true" )
	self.bHighlightReadyCol = ( Settings.ReadyColHighlight == "true" )
	self.tracklistBox:EnableCharColumn( self.bShowReadyChars )

	-- initialize list items from song database
	if (librarySize ~= 0 and not SongDB.Songs[1].Realnames) then
		
		for i = 1, #SongDB.Directories do
			local dirItem = Turbine.UI.Label();
			local _, dirLevel = string.gsub(SongDB.Directories[i], "/", "/");
			if (dirLevel == 2) then				
				dirItem:SetText(string.sub(SongDB.Directories[i],2));			
				dirItem:SetTextAlignment( Turbine.UI.ContentAlignment.MiddleLeft );
				dirItem:SetSize( 1000, 20 );				
				self.dirlistBox:AddItem( dirItem );
			end
		end
		
		self.listFrame.heading:SetText( Strings["ui_dirs"] .. " (" .. selectedDir .. ")" );
		
		if (self.dirlistBox:ContainsItem(1)) then
			local dirItem = self.dirlistBox:GetItem(1);
			dirItem:SetForeColor( self.colourDefaultHighlighted );
		end
		
		-- load content to song list box
		self:LoadSongs();
		
		-- set first item as initial selection
		local found = self.songlistBox:GetItemCount();		
		if (found > 0) then
			self.songlistBox:SetSelectedIndex( 1 )
			self:SelectSong( 1 );
		else
			self:ClearSongState( )
		end
		self.separator1.heading:SetText( Strings["ui_songs"] .. " (" .. found .. ")" );
		
		-- action for selecting a dir
		self.dirlistBox.SelectedIndexChanged = function( sender, args )
			self:SelectDir(sender:GetSelectedIndex());
		end
		-- action for selecting a song
		self.songlistBox.SelectedIndexChanged = function( sender, args )
			self:SelectSong(sender:GetSelectedIndex());
		end
		-- action for selecting a track
		self.tracklistBox.SelectedIndexChanged = function( sender, args )
			self:SelectTrack(sender:GetSelectedIndex());
		end
		
		self.tracklistBox.MouseClick = function( sender, args )
			if args.Button == Turbine.UI.MouseButton.Right then
				self:RealignTracknames( )
			end
		end
		self.sepSongsTracks.MouseClick = self.tracklistBox.MouseClick
		
	else
		-- show message when library is empty or database format has changed
		self.separator1:SetVisible( false );
		self.sepSongsTracks:SetVisible( false );
		self.listFrame.heading:SetText( "" );
		self.emptyLabel = Turbine.UI.Label();
		self.emptyLabel:SetParent( self );
		self.emptyLabel:SetPosition( 30, 155 );
		self.emptyLabel:SetSize(220, 240);
		self.emptyLabel:SetText(Strings["err_nosongs"]);
	end
	
	-- window resize control
	self.resizeCtrl = Turbine.UI.Control();
	self.resizeCtrl:SetParent(self);
	self.resizeCtrl:SetSize(20,20);		
	self.resizeCtrl:SetZOrder(200);
	self.resizeCtrl:SetPosition(self:GetWidth() - 20,self:GetHeight() - 20); 
                    
	self.resizeCtrl.MouseDown = function(sender,args)
	  sender.dragStartX = args.X;
	  sender.dragStartY = args.Y;
	  sender.dragging = true;
	end
		
	self.resizeCtrl.MouseUp = function(sender,args)
	  sender.dragging = false;
	  Settings.WindowPosition.Width = self:GetWidth();
	  Settings.WindowPosition.Height = self:GetHeight();
	end
					
	self.resizeCtrl.MouseMove = function(sender,args)
	  	local width, height = self:GetSize();
      
		if sender.dragging then
			width = width + args.X - sender.dragStartX;
			height = height + args.Y - sender.dragStartY;
			if width < self.minWidth then width = self.minWidth; end
			if height < 45 then height = 45; end
			local listContainerHeight = height - self.lCYmod;
			local tracksHeight = 0;
			if Settings.TracksVisible == "yes" then tracksHeight = Settings.TracksHeight; end
			if listContainerHeight < Settings.DirHeight + 13 + tracksHeight + 13 + 40 then
				listContainerHeight = Settings.DirHeight + 13 + tracksHeight + 13 + 40;
				height = listContainerHeight + self.lCYmod;
			end
			self:SetSize( width, height );

			self:ResizeAll( );
		end
    	
		sender:SetPosition( self:GetWidth() - sender:GetWidth(), self:GetHeight() - sender:GetHeight() );
	end -- resizeCtrl.MouseMove
	
	-- dir list, song list ratio adjust
	self.separator1.MouseDown = function(sender,args)
	  sender.dragStartX = args.X;
	  sender.dragStartY = args.Y;
	  sender.dragging = true;
	end
	
	self.separator1.MouseUp = function(sender,args)
	  sender.dragging = false;
	  Settings.DirHeight = self.dirlistBox:GetHeight(); 
	end
	
	self.separator1.MouseMove = function(sender,args)
		if ( sender.dragging ) then
			local y = self.separator1:GetTop();
			local h = self.dirlistBox:GetHeight() + args.Y - sender.dragStartY;
			if h < 40 then h = 40; end
			self.dirlistBox:SetHeight( h );
			self:AdjustSonglistHeight( )
			if self.songlistBox:GetHeight() < 40 then
				self:SetSonglistHeight( 40 )
				if (Settings.TracksVisible == "yes") then
					self.dirlistBox:SetHeight(self.listContainer:GetHeight() - Settings.TracksHeight - self.songlistBox:GetHeight() - 26);
				else
					self.dirlistBox:SetHeight(self.listContainer:GetHeight() - self.songlistBox:GetHeight() - 13);
				end
			end
			--self.separator1:SetTop(self.dirlistBox:GetHeight());
			self:SetSonglistTop( self.dirlistBox:GetHeight() + 13 )

			self:AdjustFilterUI( );
		end	
	end
	
	-- song list, track list ratio adjust
	self.sepSongsTracks.MouseDown = function(sender,args)
	  sender.dragStartX = args.X;
	  sender.dragStartY = args.Y;
	  sender.dragging = true;
	end
	
	self.sepSongsTracks.MouseUp = function(sender,args)
	  sender.dragging = false;
	  Settings.TracksHeight = self.tracklistBox:GetHeight(); 
	end
	
	self.sepSongsTracks.MouseMove = function(sender,args)
		if ( sender.dragging ) then
			local y = self.sepSongsTracks:GetTop();
			local h = self.tracklistBox:GetHeight() - args.Y + sender.dragStartY;
			if h < 40 then h = 40; end
			self:SetTracklistHeight( h )
			self:UpdateTracklistTop( );
			self:AdjustSonglistHeight( )
		end
		if self.songlistBox:GetHeight() < 40 then
			self:SetSonglistHeight( 40 )
			self:SetTracklistHeight( self.listContainer:GetHeight() - self.dirlistBox:GetHeight() - self.songlistBox:GetHeight() - 26 );
			self:UpdateTracklistTop( );
		end	
	end
	
	-- Settings button	
	self.settingsBtn = Turbine.UI.Lotro.Button();
	self.settingsBtn:SetParent( self );
	self.settingsBtn:SetPosition(self:GetWidth()/2 - 55, self:GetHeight() - 30 );	
	self.settingsBtn:SetSize(110,20);
	self.settingsBtn:SetText(Strings["ui_settings"]);
	
	-- actions for settings button
	self.settingsBtn.MouseClick = function(sender, args)
		if(args.Button == Turbine.UI.MouseButton.Left) then
			settingsWindow:SetVisible(true);
		end
	end	
	
	-- action for closing window and saving position
	self.Closed = function( sender, args )
		self:SaveSettings();
		self:SetVisible( false );
	end
	
	if (Plugins ["Songbook"] ~= nil ) then
		Plugins["Songbook"].Unload = function( sender, args )
			self:SaveSettings();
			RemoveCallback( Turbine.Chat, "Received", ChatHandler );
		end
	end
	
	self:CreateTimerUI( )
	self:CreateFilterUI( ) -- Creates the UI elements for the filters
	AddCallback( Turbine.Chat, "Received", ChatHandler ); -- installs handler for chat messages (to catch ready messages)
	self.listboxPlayers:EnableCharColumn( self.bShowReadyChars )
	self:RefreshPlayerListbox( ) -- lists the current party members; more will be added through chat messages
	
	if Settings.FiltersState == "true" then self:ShowFilterUI( true ); end
	self:SetChiefMode( Settings.ChiefMode == "true" )
	self:HightlightReadyColumns( self.bHighlightReadyCol )
	
	self:ResizeAll( ); -- Adjust variable sizes and positions to current main window size
end -- SongbookWindow:Constructor()

-- TODO: add complete set of checks
function SongbookWindow:ValidateWindowPosition( winPos )
	if winPos.Left < 0 or winPos.Top < 0 or winPos.Width < 0 or winPos.Height < 0 then
		winPos.Left = 700
		winPos.Top = 20
		winPos.Width = 342
		winPos.Height = 398
	end
end


function SongbookWindow:ActivateTimer( bActivate )
	self.bTimer = bActivate
	if not bActivate then self:StopTimer( ); end
end

function SongbookWindow:CreateTimerUI( )
	self.tracksMsg = Turbine.UI.Label( )
	self.tracksMsg:SetParent( self.listContainer )
	self.tracksMsg:SetMultiline( false )
	self.tracksMsg:SetSize( 150, 20 )
	self.tracksMsg:SetTextAlignment( Turbine.UI.ContentAlignment.MiddleRight )
	self.tracksMsg:SetZOrder( 350 )
	self.tracksMsg:SetVisible( false )

	self.Update =
		function( )
			local previous = self.currentTime
			self.currentTime = Turbine.Engine.GetLocalTime() - self.startTimer
			local bKeepRunning = ( self.songDuration == 0 or self.currentTime <= self.songDuration )
			if self.bTimerCountdown == true and self.songDuration > 0 then
				self.currentTime = self.songDuration - self.currentTime;
				bKeepRunning = self.currentTime >= 0
			end

			if bKeepRunning then
				if self.currentTime ~= previous then self:SetTimer( self.currentTime ); end
			else
				self:StopTimer( )
			end
		end
end

function SongbookWindow:SetTimer( value )
	local mins = math.floor( value / 60 )
	self.tracksMsg:SetText( string.format("%u:%02u", mins, value - mins * 60 ) )
end
	
function SongbookWindow:StartTimer( )
	self.startTimer = Turbine.Engine.GetLocalTime( )
	self.songDuration = 0
	local item, songTime
	for i = 1, self.tracklistBox:GetItemCount( ) do
		item = self.tracklistBox:GetItem( i )
		sMinutes, sSeconds = string.match( item:GetText( ), ".*%((%d+):(%d+)%).*" ) -- try (mm:ss)
		if not sMinutes or not sSeconds then
			sMinutes, sSeconds = string.match( item:GetText( ), ".*(%d+):(%d+).*" ) -- no luck, try just mm:ss
		end
		if sMinutes and sSeconds and tonumber(sMinutes) < 60 and tonumber(sSeconds) < 60 then
			songTime = sMinutes * 60 + sSeconds
			if songTime > self.songDuration then self.songDuration = songTime; end -- need longest track
		end
	end

	if self.bTimerCountdown == true and self.songDuration > 0 then self.currentTime = self.songDuration
	else self.currentTime = 0; end

	self:SetTimer( self.currentTime )
	self.tracksMsg:SetVisible( true )
	self:SetWantsUpdates( true )
end

function SongbookWindow:StopTimer( )
	self:SetWantsUpdates( false )
	self.tracksMsg:SetVisible( false )
end

function SongbookWindow:ResizeAll( )
	self.listFrame:SetSize(self:GetWidth() - self.lFXmod, self:GetHeight() - self.lFYmod);
	self.listContainer:SetSize(self:GetWidth() - self.lCXmod, self:GetHeight() - self.lCYmod);
	self.listFrame.heading:SetSize(self.listFrame:GetWidth(),13);
	self.instrContainer:SetTop( self:GetHeight() - 75 );

	self.dirlistBox:SetHeight( Settings.DirHeight );
	self:AdjustDirlistPosition( );

	if Settings.TracksVisible == "yes" then
		self:AdjustTracklistSize( Settings.TracksHeight )
		self:UpdateTracklistTop( )
	else
		self.tracksMsg:SetPosition( self.dirlistBox:GetLeft()+self.dirlistBox:GetWidth()-160, self.dirlistBox:GetTop()+self.dirlistBox:GetHeight() );	
	end

	self:AdjustSonglistHeight( );
	self:AdjustSonglistPosition( );
		
	self.songTitle:SetWidth(self:GetWidth() - 52);
	self.settingsBtn:SetPosition(self:GetWidth()/2 - 55, self:GetHeight() - 30 );	
	--self.cbFilters:SetPosition( self:GetWidth()/2 + 65, self:GetHeight() - 30 );
	self.tipLabel:SetLeft(self:GetWidth() - 270);
	self:AdjustFilterUI( );
end -- SongbookWindow:ResizeAll( )


-- action for selecting a directory
function SongbookWindow:SelectDir( iDir )
	local selectedItem = self:SetListboxColours( self.dirlistBox ) --, iDir )
	if not selectedItem then return; end
	if selectedItem:GetText() == ".." then
		selectedDir = "";
		table.remove(dirPath,#dirPath);
		for i = 1,#dirPath do 
			selectedDir = selectedDir .. dirPath[i];
		end
	else		
		selectedDir = selectedDir .. selectedItem:GetText();
		dirPath[#dirPath+1] = selectedItem:GetText();	
	end
	
	if (string.len(selectedDir)<31) then 
		self.listFrame.heading:SetText( Strings["ui_dirs"] .. " (" .. selectedDir .. ")" );
	else 
		self.listFrame.heading:SetText( Strings["ui_dirs"] .. " (" .. string.sub(selectedDir,string.len(selectedDir)-30) .. ")" );
	end
	
	-- refresh dir list
	self.dirlistBox:ClearItems();
	local dirItem = Turbine.UI.Label();
	if (selectedDir ~= "/") then
		dirItem:SetText(".."); -- first item as link to previous directory
		dirItem:SetTextAlignment( Turbine.UI.ContentAlignment.MiddleLeft );
		dirItem:SetSize( 1000, 20 );				
		self.dirlistBox:AddItem( dirItem );
	end
	
	for i = 1, #SongDB.Directories do
		dirItem = Turbine.UI.Label();		
		local _, dirLevelIni = string.gsub(selectedDir, "/", "/");
		local _, dirLevel = string.gsub(SongDB.Directories[i], "/", "/");
		if (dirLevel == dirLevelIni + 1) then
			if (selectedDir ~= "/") then
				local matchPos,_ = string.find(SongDB.Directories[i], selectedDir, 0, true);
				if (matchPos == 1) then	
					local _,cutPoint = string.find(SongDB.Directories[i], dirPath[#dirPath], 0, true);
					dirItem:SetText(string.sub(SongDB.Directories[i],cutPoint+1));			
					dirItem:SetTextAlignment( Turbine.UI.ContentAlignment.MiddleLeft );
					dirItem:SetSize( 1000, 20 );				
					self.dirlistBox:AddItem( dirItem );
				end
			else 
				dirItem:SetText(string.sub(SongDB.Directories[i],2));			
				dirItem:SetTextAlignment( Turbine.UI.ContentAlignment.MiddleLeft );
				dirItem:SetSize( 1000, 20 );				
				self.dirlistBox:AddItem( dirItem );
			end
		end
	end
	
	self.songlistBox:ClearItems();
	self:LoadSongs( );
	self:InitSonglist( );
end -- SelectDir


-- load content to song list box
function SongbookWindow:LoadSongs( )
	local nFiltered = 0;
	for i = 1, librarySize do
		local songItem = Turbine.UI.Label();
		-- Added function to filter song data
		if( SongDB.Songs[i].Filepath == selectedDir and self:ApplyFilters( SongDB.Songs[i] ) ) then
			if (Settings.DescriptionVisible == "yes") then
				if (Settings.DescriptionFirst == "yes") then
					songItem:SetText(SongDB.Songs[i].Tracks[1].Name .. " / " .. SongDB.Songs[i].Filename);					
				else
					songItem:SetText(SongDB.Songs[i].Filename .. " / " .. SongDB.Songs[i].Tracks[1].Name);
				end
			else
				songItem:SetText( SongDB.Songs[i].Filename );
			end
			songItem:SetTextAlignment( Turbine.UI.ContentAlignment.MiddleLeft );
			songItem:SetSize( 1000, 20 );				
			self.songlistBox:AddItem( songItem );
			nFiltered = nFiltered + 1;
			self.aFilteredIndices[ nFiltered ] = i; -- Create filtered index
		end
	end
end

function SongbookWindow:SelectedTrackIndex( iList )
	if not iList then iList = selectedTrack; end -- use global selected track index if none provided
	if self.iCurrentSetup and self.aSetupTracksIndices[ iList ] then
		return self.aSetupTracksIndices[ iList ];
	end
	return iList;
end

-- action for selecting a song
function SongbookWindow:SelectSong( iSong )
	if( iSong < 1 or iSong > self.songlistBox:GetItemCount( ) ) then
		return;
	end
	selectedTrack = 1;
	self.aSetupTracksIndices = { };
	self.aSetupListIndices = { };
	self.iCurrentSetup = nil

	-- clear focus
	self:SetListboxColours( self.songlistBox ) --, iSong )
	
	selectedSongIndex = self.aFilteredIndices[ iSong ];
	selectedSong = SongDB.Songs[selectedSongIndex].Filename;
			
	if ( SongDB.Songs[selectedSongIndex].Tracks[1].Name ~= "") then
		self.songTitle:SetText( SongDB.Songs[selectedSongIndex].Tracks[1].Name );	
	else
		self.songTitle:SetText( SongDB.Songs[selectedSongIndex].Filename );	
	end
	self.trackNumber:SetText( SongDB.Songs[selectedSongIndex].Tracks[1].Id );
	self.trackPrev:SetVisible(false);
	
	if (#SongDB.Songs[selectedSongIndex].Tracks > 1) then
		self.trackNext:SetVisible(true);
	else
		self.trackNext:SetVisible(false);
	end

	self:ListTracks(selectedSongIndex);	
	
	self:ClearPlayerReadyStates( );
	self:SelectTrack( selectedTrack );
	self:SetPlayerColours( );
	self:ListSetups( selectedSongIndex )
	self.iCurrentSetup = self:SetupIndexForCount( selectedSongIndex, self.selectedSetupCount )
	self:SelectSetup( self.iCurrentSetup )
	self:UpdateSetupColours( );
	
	local found = self.tracklistBox:GetItemCount();
	self.sepSongsTracks.heading:SetText( Strings["ui_parts"] .. " (" .. found .. ")" );
end


-- action for repopulating the track list when song is changed
function SongbookWindow:ListTracks( songid )			
	self.tracklistBox:ClearItems();
	for i = 1, #SongDB.Songs[songid].Tracks do
		self:AddTrackToList( songid, i )
	end
	--Turbine.Chat.Received = self.ChatHandler; -- Enable chat monitoring for ready messages to update track colours
end


function SongbookWindow:CreateTracklistItem( sText )
	local trackItem = Turbine.UI.Label();
	trackItem:SetMultiline( false )
	trackItem:SetText( sText );
	trackItem.MouseClick = self.sepSongsTracks.MouseClick
	trackItem:SetForeColor( self.colourDefault );
	return trackItem
end

function SongbookWindow:AddTrackToList( iSong, iTrack )
	local sTerseName = self:TerseTrackname( SongDB.Songs[iSong].Tracks[iTrack].Name );
	local trackItem = self:CreateTracklistItem( "[" .. SongDB.Songs[iSong].Tracks[iTrack].Id .. "] " .. sTerseName )
	trackItem:SetTextAlignment( Turbine.UI.ContentAlignment.MiddleLeft );
	trackItem:SetSize( 1000, 20 );
	self.tracklistBox:AddItem( trackItem );			
end

	
-- Right-align track names (so user can quickly check then end of the track name)
function SongbookWindow:RealignTracknames( )
	local alignment, left
	if self.alignTracksRight == false then
		self.alignTracksRight = true
		alignment = Turbine.UI.ContentAlignment.MiddleRight
		left = self.tracklistBox:GetWidth( ) - 1010
	else
		self.alignTracksRight = false
		alignment = Turbine.UI.ContentAlignment.MiddleLeft
		if self.bShowReadyChars then left = 20
		else left = 0; end
	end

	for i = 1, self.tracklistBox:GetItemCount( ) do
		local item = self.tracklistBox:GetItem( i )
		item:SetLeft( left ) 
		item:SetTextAlignment( alignment );
	end
end

-- action for changing track selection (trackid is listbox index)
function SongbookWindow:SelectTrack( trackid )
	if self.bShowReadyChars then trackid = math.floor( (trackid+1) / 2 ); end
	selectedTrack = trackid;
	local iTrack = self:SelectedTrackIndex( trackid );
	local trackcount = #SongDB.Songs[selectedSongIndex].Tracks;
--Turbine.Shell.WriteLine( "Selected: " .. tostring(trackid) .. " Index: " .. tostring(iTrack) );

	if selectedTrack > 1 then
		if selectedTrack == trackcount then
			self.trackPrev:SetVisible( true );
			self.trackNext:SetVisible( false );
		else
			self.trackPrev:SetVisible( true );
			self.trackNext:SetVisible( true );
		end
	end
	if ( selectedTrack == 1) then
		self.trackPrev:SetVisible( false );
		if (trackcount == 1) then		
			self.trackNext:SetVisible( false );
		else
			self.trackNext:SetVisible( true );
		end
	end

	self.trackNumber:SetText(SongDB.Songs[selectedSongIndex].Tracks[iTrack].Id);
	self.songTitle:SetText(SongDB.Songs[selectedSongIndex].Tracks[iTrack].Name);

	self.playSlotShortcut = Turbine.UI.Lotro.Shortcut( Turbine.UI.Lotro.ShortcutType.Alias, Strings["cmd_play"] .. " \"" .. SongDB.Songs[selectedSongIndex].Filepath .. selectedSong .. "\" " .. SongDB.Songs[selectedSongIndex].Tracks[iTrack].Id);
	self.playSlot:SetShortcut( self.playSlotShortcut );
	self.playSlot:SetVisible( true );

	self.syncSlotShortcut = Turbine.UI.Lotro.Shortcut( Turbine.UI.Lotro.ShortcutType.Alias, Strings["cmd_play"] .. " \"" .. SongDB.Songs[selectedSongIndex].Filepath .. selectedSong .. "\" " .. SongDB.Songs[selectedSongIndex].Tracks[iTrack].Id .. " " .. Strings["cmd_sync"]);
	self.syncSlot:SetShortcut( self.syncSlotShortcut );
	self.syncSlot:SetVisible( true );
	
	self.shareSlot:SetShortcut( Turbine.UI.Lotro.Shortcut( Turbine.UI.Lotro.ShortcutType.Alias, self:ExpandCmd(Settings.DefaultCommand)));		
	self.shareSlot:SetVisible( true );		

	self:SetTrackColours( selectedTrack );
end

-- action for setting focus on the track list
function SongbookWindow:SetTrackColours( iSelectedTrack )
	if not self.tracklistBox or self.tracklistBox:GetItemCount( ) < 1 then return; end
	self:ClearPlayerReadyStates( ); -- Clear ready states for currently displayed song

	for iTrack = 1,#SongDB.Songs[selectedSongIndex].Tracks do
		if self.iCurrentSetup and not self.aSetupListIndices[ iTrack ] then
			self:GetTrackReadyState( SongDB.Songs[selectedSongIndex].Tracks[ iTrack ].Name, 3 );
		else
			local iList = iTrack;
			if self.aSetupListIndices[ iTrack ] then iList = self.aSetupListIndices[ iTrack ]; end
			local item = self.tracklistBox:GetItem(iList);
			local readyState = self:GetTrackReadyState( SongDB.Songs[selectedSongIndex].Tracks[iTrack].Name );			
			item:SetForeColor( self:GetColourForTrack( readyState, iList == iSelectedTrack ) );
			item:SetBackColor( self:GetBackColourForTrack( readyState ) );
			self:SetTrackReadyChar( iList, readyState )
		end
	end
end

-- return track colour based on readyState retrieved from GetTrackReadyState(...)
function SongbookWindow:GetColourForTrack( readyState, bSelectedTrack )
	if bSelectedTrack then
		if not readyState then -- track not ready
			return self.colourDefaultHighlighted;
		elseif readyState == 0 then -- track is ready by more than one player
			return self.colourReadyMultipleHighlighted;
		else -- track ready by one player
			return self.colourReadyHighlighted;
		end
	else
		if not readyState then
			return self.colourDefault;
		elseif readyState == 0 then
			return self.colourReadyMultiple;
		else
			return self.colourReady;
		end
	end
end

-- background colour indicates the track one has ready
-- TODO: blue (multiple ready track) overrides wrong instrument indicator
function SongbookWindow:GetBackColourForTrack( readyState )
	if not readyState or readyState ~= self.sPlayerName then return self.backColourDefault; end
	return self.backColourHighlight
	--if self.bInstrumentOk then return self.backColourHighlight; end
	--return self.backColourWrongInstrument
end	

-- set track ready indicator
function SongbookWindow:SetTrackReadyChar( iList, readyState )
	if not readyState then -- track not ready
		self.tracklistBox:SetColumnChar( iList, self.chNone, false )
	elseif readyState == 0 then -- track is ready by more than one player
		self.tracklistBox:SetColumnChar( iList, self.chMultiple, true )
	else -- track ready by one player
		self.tracklistBox:SetColumnChar( iList, self.chReady, false )
	end
end


-- action to search songs
function SongbookWindow:SearchSongs()
	self.songlistBox:ClearItems()
	local matchFound
	local nFound = 0
	
	self.aFilteredIndices = { }
	for i = 1, librarySize do		
		matchFound = false

		if self:ApplyFilters( SongDB.Songs[i] ) == true then -- filters are matched, now look for search input
			if string.find( string.lower(SongDB.Songs[i].Filename ), string.lower( self.searchInput:GetText( ) ) ) ~= nil then
				matchFound = true;
			else
				for j = 1, #SongDB.Songs[i].Tracks do
					if string.find(string.lower(SongDB.Songs[i].Tracks[j].Name), string.lower(self.searchInput:GetText())) ~= nil then
						matchFound = true;
						break;
					end
				end
			end
		end
		
		if matchFound == true then
			local songItem = Turbine.UI.Label();
			if (Settings.DescriptionVisible == "yes") then			
				songItem:SetText(SongDB.Songs[i].Filename .. " / " .. SongDB.Songs[i].Tracks[1].Name);
			else
				songItem:SetText(SongDB.Songs[i].Filename);
			end
			songItem:SetTextAlignment( Turbine.UI.ContentAlignment.MiddleLeft );
			songItem:SetSize( 1000, 20 );				
			self.songlistBox:AddItem( songItem );
			nFound = nFound + 1;
			self.aFilteredIndices[ nFound ] = i; -- Create index redirect table
		end
	end
	
	--local found = self.songlistBox:GetItemCount();
	if (nFound > 0) then self:SelectSong(1);
	else self:ClearSongState( );  end
	self.separator1.heading:SetText( Strings["ui_songs"] .. " (" .. nFound .. ")" );
end

-- action for toggling search function on and off
function SongbookWindow:ToggleSearch(mode)
	if (Settings.SearchVisible == "yes" or mode == "off") then		
		Settings.SearchVisible = "no";
		self:SetSearch( -20, false )
	else
		Settings.SearchVisible = "yes";
		self:SetSearch( 20, true )
	end

	self.listFrame:SetHeight(self:GetHeight() - self.lFYmod);	
	self.listContainer:SetHeight(self:GetHeight() - self.lCYmod);		
	if (Settings.TracksVisible == "no") then
		self:ShowTrackListbox( false ) -- ?
	end
end

function SongbookWindow:SetSearch( delta, bShow )
	self.searchInput:SetVisible(bShow);
	self.searchBtn:SetVisible(bShow);
	self.clearBtn:SetVisible(bShow);		
	self.lFYmod = self.lFYmod + delta;		
	self.lCYmod = self.lCYmod + delta;
	self.listFrame:SetTop( self.listFrame:GetTop( ) + delta);
	self.listContainer:SetTop( self.listContainer:GetTop( ) + delta );
	self:SetSonglistHeight(self.songlistBox:GetHeight() - delta);		
	self:MoveTracklistTop( -delta )
end

-- action for toggling description on and off
function SongbookWindow:ToggleDescription()
	if (Settings.DescriptionVisible == "yes") then
		Settings.DescriptionVisible = "no";
		self.songlistBox:ClearItems();
		self:LoadSongs();
		local found = self.songlistBox:GetItemCount();		
		if (found > 0) then self:SelectSong(1);
		else self:ClearSongState( ); end
	else
		Settings.DescriptionVisible = "yes";
		self.songlistBox:ClearItems();
		self:LoadSongs();
		local found = self.songlistBox:GetItemCount();		
		if (found > 0) then self:SelectSong(1);
		else self:ClearSongState( ); end
	end
end

-- action for toggling description on and off
function SongbookWindow:ToggleDescriptionFirst()
	if (Settings.DescriptionFirst == "yes") then
		Settings.DescriptionFirst = "no";		
		if (Settings.DescriptionVisible == "yes") then
			self.songlistBox:ClearItems();
			self:LoadSongs();
			local found = self.songlistBox:GetItemCount();		
			if (found > 0) then self:SelectSong(1);
			else self:ClearSongState( ); end
		end
	else
		Settings.DescriptionFirst = "yes";
		if (Settings.DescriptionVisible == "yes") then
			self.songlistBox:ClearItems();
			self:LoadSongs();
			local found = self.songlistBox:GetItemCount();		
			if (found > 0) then self:SelectSong(1);
			else self:ClearSongState( ); end
		end
	end
end

-- action for toggling tracks display on and off
function SongbookWindow:ToggleTracks()
	if (Settings.TracksVisible == "yes") then
		Settings.TracksVisible = "no";
		self:SetSonglistHeight(self.listContainer:GetHeight() - self.dirlistBox:GetHeight() - 13);
		self:ShowTrackListbox( false )
		self.listboxSetups:SetVisible(false);
		self.tracksMsg:SetPosition( self.dirlistBox:GetLeft()+self.dirlistBox:GetWidth()-150, self.dirlistBox:GetTop()+self.dirlistBox:GetHeight() );	
	else
		Settings.TracksVisible = "yes";
		self:ShowTrackListbox( true )
		self.listboxSetups:SetVisible(self.bShowSetups)
		
		-- check if there's room for the track list and adjust
		local h = self.dirlistBox:GetHeight() + Settings.TracksHeight + 26;
		if (self.listContainer:GetHeight() - h < 40) then
			self.listContainer:SetHeight(h + self.songlistBox:GetHeight())
			self:SetHeight(self.listContainer:GetHeight() + self.lCYmod);
			self.listFrame:SetHeight(self:GetHeight() - self.lFYmod);
			self.resizeCtrl:SetTop(self:GetHeight() - 20); 
		end
					
		self:SetTracklistTop( self.listContainer:GetHeight() - Settings.TracksHeight )
		self:AdjustTracklistSize( Settings.TracksHeight )			
		self:SetSonglistHeight(self.listContainer:GetHeight() - self.dirlistBox:GetHeight() - self.tracklistBox:GetHeight() - 26);
		self.settingsBtn:SetPosition(self:GetWidth()/2 - 55, self:GetHeight() - 30 );	
		--self.cbFilters:SetPosition( self:GetWidth()/2 + 65, self:GetHeight() - 30 );
	end
end

-- action for toggling instrument slots on and off
function SongbookWindow:ToggleInstrSlots()
	local hMod = 45;
	if (CharSettings.InstrSlots["visible"] == "yes") then		
		CharSettings.InstrSlots["visible"] = "no";		
		self.instrContainer:SetVisible( false );	
		self:SetInstrSlots( -hMod )
	else
		CharSettings.InstrSlots["visible"] = "yes";
		self:SetInstrSlots( hMod )
		self.instrContainer:SetVisible( true );
	end
end

function SongbookWindow:SetInstrSlots( delta )
	self.lFYmod = self.lFYmod + delta;
	self.lCYmod = self.lCYmod + delta;
	self.listFrame:SetHeight(self.listFrame:GetHeight() - delta);
	self.listContainer:SetHeight(self.listContainer:GetHeight() - delta);				
	self:SetSonglistHeight(self.songlistBox:GetHeight() - delta);
	if (Settings.TracksVisible == "yes") then
		self:MoveTracklistTop( -delta )
		--self.tracklistBox:SetTop(self.tracklistBox:GetTop() - hMod);
		--self.sepSongsTracks:SetTop(self.sepSongsTracks:GetTop() - hMod);		
	end
end

function SongbookWindow:ClearSlots()
	for i=1,CharSettings.InstrSlots["number"] do
		CharSettings.InstrSlots[tostring(i)].qsType ="";
		CharSettings.InstrSlots[tostring(i)].qsData = "";
		local sc = Turbine.UI.Lotro.Shortcut( "", "");
		self.instrSlot[i]:SetShortcut(sc);
	end
end

function SongbookWindow:AddSlot()
	if self:GetWidth() > 10+(CharSettings.InstrSlots["number"]+1)*40 then
		local newslot = CharSettings.InstrSlots["number"]+1;
		CharSettings.InstrSlots["number"] = newslot;
		self.instrSlot[newslot] = Turbine.UI.Lotro.Quickslot();
		self.instrSlot[newslot]:SetParent( self.instrContainer );
		self.instrSlot[newslot]:SetPosition(40*(newslot-1), 0);
		self.instrSlot[newslot]:SetSize(37, 37);
		self.instrSlot[newslot]:SetZOrder(100);
		self.instrSlot[newslot]:SetAllowDrop(true);
		self.instrContainer:SetWidth(self.instrContainer:GetWidth()+40); 
		
		local sc = Turbine.UI.Lotro.Shortcut("","");
		self.instrSlot[newslot]:SetShortcut(sc);
		
		CharSettings.InstrSlots[tostring(newslot)] = { qsType = "", qsData = "" };		
		
		self.instrSlot[newslot].ShortcutChanged = function( sender, args )
			pcall(function() 
				local sc = sender:GetShortcut();
				CharSettings.InstrSlots[tostring(newslot)].qsType = tostring(sc:GetType());
				CharSettings.InstrSlots[tostring(newslot)].qsData = sc:GetData();
			end);
		end
		
		self.instrSlot[newslot].DragLeave = function( sender, args )
			if (instrdrag) then 
				CharSettings.InstrSlots[tostring(newslot)].qsType ="";
				CharSettings.InstrSlots[tostring(newslot)].qsData = "";
				local sc = Turbine.UI.Lotro.Shortcut( "", "");
				self.instrSlot[newslot]:SetShortcut(sc);
				instrdrag = false;
			end
		end
		
		self.instrSlot[newslot].MouseDown = function( sender, args )
			if(args.Button == Turbine.UI.MouseButton.Left) then	
				instrdrag = true;
			end
		end
	end
end

function SongbookWindow:DelSlot()
	if CharSettings.InstrSlots["number"] > 1 then
		local delslot = CharSettings.InstrSlots["number"];
		CharSettings.InstrSlots["number"] = CharSettings.InstrSlots["number"]-1;
		self.instrContainer:SetWidth(self.instrContainer:GetWidth()-40); 
		self.instrSlot[delslot] = nil;
		CharSettings.InstrSlots[tostring(delslot)] = nil;
	end
end


function SongbookWindow:ExpandCmd(cmdId)
	local selTrack = self:SelectedTrackIndex( );
	if librarySize ~= 0 then
		local cmd = Settings.Commands[cmdId].Command;
		if SongDB.Songs[selectedSongIndex].Tracks[ selTrack ] then
			cmd = string.gsub(cmd, "%%name", SongDB.Songs[selectedSongIndex].Tracks[ selTrack ].Name);				
			cmd = string.gsub(cmd, "%%file", SongDB.Songs[selectedSongIndex].Filename);
			if (selTrack ~= 1) then
				cmd = string.gsub(cmd, "%%part", selTrack );
			else
				cmd = string.gsub(cmd, "%%part", "");
			end
		elseif SongDB.Songs[selectedSongIndex].Filename then
			cmd = string.gsub(cmd, "%%name", SongDB.Songs[selectedSongIndex].Filename);
			cmd = string.gsub(cmd, "%%file", SongDB.Songs[selectedSongIndex].Filename);	
			if (selTrack ~= 1) then
				cmd = string.gsub(cmd, "%%part", selTrack);
			else
				cmd = string.gsub(cmd, "%%part", "");
			end
		else 
			cmd = "";
		end
		return cmd;
	end
end

function SongbookWindow:SaveSettings()
	Settings.WindowPosition.Left = tostring(self:GetLeft());
	Settings.WindowPosition.Top = tostring(self:GetTop());
	Settings.WindowPosition.Width = tostring(self:GetWidth());
	Settings.WindowPosition.Height = tostring(self:GetHeight());
	Settings.ToggleTop = tostring(Settings.ToggleTop);
	Settings.ToggleLeft = tostring(Settings.ToggleLeft);
	Settings.DirHeight = tostring(Settings.DirHeight);
	Settings.TracksHeight = tostring(Settings.TracksHeight);
	Settings.WindowOpacity = tostring(Settings.WindowOpacity);
	Settings.ToggleOpacity = tostring(Settings.ToggleOpacity);
	Settings.FiltersState = tostring( self.bFilter )
	Settings.ChiefMode = tostring( self.bChiefMode )
	Settings.TimerState = tostring( self.bTimer )
	Settings.TimerCountdown = tostring( self.bTimerCountdown )
	Settings.ReadyColState = tostring( self.bShowReadyChars )
	Settings.ReadyColHighlight = tostring( self.bHighlightReadyCol )
	
	for i = 1, CharSettings.InstrSlots["number"] do
		CharSettings.InstrSlots[tostring(i)].qsType = tostring(CharSettings.InstrSlots[tostring(i)].qsType);
	end
	CharSettings.InstrSlots["number"] = tostring(CharSettings.InstrSlots["number"]);
	
	SongbookSave( Turbine.DataScope.Account, gSettings, Settings,
		function( result, message )
			if ( result ) then
				Turbine.Shell.WriteLine( "<rgb=#00FF00>" .. Strings["sh_saved"] .. "</rgb>");
			else
				Turbine.Shell.WriteLine( "<rgb=#FF0000>" .. Strings["sh_notsaved"] .. " " .. message .. "</rgb>" );
			end
		end);
	SongbookSave( Turbine.DataScope.Character, gSettings, CharSettings,
		function( result, message )
			if ( result ) then
				--Turbine.Shell.WriteLine( "<rgb=#00FF00>" .. Strings["sh_saved"] .. "</rgb>");
			else
				Turbine.Shell.WriteLine( "<rgb=#FF0000>" .. Strings["sh_notsaved"] .. " " .. message .. "</rgb>" );
			end
		end);	
end


-- Parse filter string entered by the user.
function SongbookWindow:ParsePartsFilter( sText )
	local sPattern = "[";
	local iEnd = 0;
	local number, numberTo, iEndTo, temp, maxTracks;

	for maxTracks = 1,self.maxTrackCount do
		iEnd = iEnd + 1;
		temp, iEnd, number = string.find( sText, "%s*(%d+)%s*", iEnd );

		if( iEnd == nil ) then break; end

		iEnd = iEnd + 1;
		if( string.sub( sText, iEnd, iEnd ) == "-" ) then
			temp, iEndTo, numberTo = string.find( sText, "%s*(%d+)%s*", iEnd + 1 );
			if( iEndTo == nil ) then
				numberTo = self.maxTrackCount;
			else
				iEnd = iEndTo + 1;
			end
		else
			numberTo = number;
		end
	
		for temp = number, numberTo do
			sPattern = sPattern .. string.char( 0x40 + temp ); -- 0x40 is ASCII-code 'A' - 1
		end
	end
	
	if( sPattern == "[" ) then
		self.sFilterPartcount = "[a-z]";
	else
		self.sFilterPartcount = sPattern .. "]";
	end
end -- ParsePartsFilter


-- return true if at least one word is in both string lists 
function SongbookWindow:MatchStringList( list1, list2 )
	for word1 in string.gmatch( list1, "%a+" ) do
		for word2 in string.gmatch( list2, "%a+" ) do
			if word1 == word2 then return true; end
		end
	end
	return false;
end

function SongbookWindow:IsEmptyString( s )
  return not not tostring( s ):find( "^%s*$" )
end

-- Check whether the given song fits all the filters that are currently set 
function SongbookWindow:ApplyFilters( songData )
	if( songData == nil ) then return false; end
	
	if( self.cbComposer and self.cbComposer:IsChecked( ) ) then
		if( songData.Artist == nil ) then return false; end
		local sFilter = string.lower( self.editComposer:GetText( ) );
		if( sFilter ~= "" and string.find( string.lower( songData.Artist ), sFilter ) == nil ) then
			return false
		end
	end
	if( self.cbPartcount and self.cbPartcount:IsChecked( ) ) then
		local sFilter = self.editPartcount:GetText( )
		if( sFilter == "" ) then
			if not self.maxPartCount then return true
			else sFilter = "1-" .. tostring( self.maxPartCount ); end
		end
		if( songData.Partcounts == nil ) then return false; end
		self:ParsePartsFilter( sFilter );
		if( string.match( songData.Partcounts, self.sFilterPartcount ) == nil ) then
			return false -- Song does not have a setup with an acceptable number of players
		end
	end
	if( self.cbGenre and self.cbGenre:IsChecked( ) ) then
		if( songData.Genre == nil ) then return false; end
		local sFilter = string.lower( self.editGenre:GetText( ) );
		if sFilter ~= "" and not self:MatchStringList( sFilter, string.lower( songData.Genre ) ) then
		--if( sFiler ~= "" and string.find( string.lower( songData.Genre ), sFilter ) == nil ) then
			return false
		end
	end
	if( self.cbMood and self.cbMood:IsChecked( ) ) then
		if( songData.Mood == nil ) then return false; end;
		local sFilter = string.lower( self.editMood:GetText( ) );
		if sFilter ~= "" and not self:MatchStringList( sFilter, string.lower( songData.Mood ) ) then
		--if( sFilter ~= "" and string.find( sFilter, string.lower( songData.Mood ) ) == nil ) then
			return false;
		end
	end
	if( self.cbAuthor and self.cbAuthor:IsChecked( ) ) then
		if( songData.Transcriber == nil ) then return false; end;
		local sFilter = string.lower( self.editAuthor:GetText( ) );
		if( sFilter ~= "" and string.find( sFilter, string.lower( songData.Transcriber ) ) == nil ) then
			return false;
		end
	end

	return true;
end -- ApplyFilters


-- Update the song list
function SongbookWindow:UpdateSongs( )
	self.songlistBox:ClearItems( );
	
	local sSearch = self.searchInput:GetText( );
	if( sSearch and sSearch ~= "" ) then self:SearchSongs( );
	else self:LoadSongs( ); end
	
	self:InitSonglist( );
end -- UpdateSongs

-- Initialize song: List tracks, set/clear chat handler, set headings
function SongbookWindow:InitSonglist( )
	local nSongs = self.songlistBox:GetItemCount( );
	if nSongs > 0 then
		self:SelectSong( 1 );
	else
		self.tracklistBox:ClearItems( );
		self:ClearSongState( )
		--Turbine.Chat.Received = nil; -- No tracks listed, so deactivate player ready indicator
		self.sepSongsTracks.heading:SetText( Strings["ui_parts"] .. " (0)" );
	end
	self.separator1.heading:SetText( Strings["ui_songs"] .. " (" .. nSongs .. ")" );
end -- UpdateSongs

-- Create compact track name by removing the title.
-- Note: Many of our older songs have quite different naming schemes; not sure if it's even worth parsing.
function SongbookWindow:TerseTrackname( sTrack )
	return sTrack; -- disabled for now.
end


function SongbookWindow:ClearSongState(  )
	self.aReadyTracks = "";
	self:ClearPlayerStates( );
	self:ClearSetups(  );
	self:SetTrackColours( selectedTrack );
	self:SetPlayerColours( );
end

function SongbookWindow:SongStarted( )
	self:ClearSongState(  )
	if self.bInstrumentOk == false then
		self.tracksMsg:SetForeColor( self.colourDefault )
		self.tracksMsg:SetVisible( false )
		self.bInstrumentOk = true
	end
	if self.bTimer then self:StartTimer( )
	else --self:StopTimer( ) -- in case it is still counting ...
	end
end

-- Handler for chat messages to indicate players readying tracks
function ChatHandler( sender, args )
	local sMessage = args.Message;

	if( args.ChatType ~= Turbine.ChatType.Standard ) then
		return; -- Player ready messages appear in the standard chat
	end
	
	if string.find( sMessage, Strings["chat_playBegin"] ) ~= nil or string.find( sMessage, Strings["chat_playBeginSelf"] ) ~= nil then
		songbookWindow:SongStarted( );
		return;
	end

	if string.find( sMessage, Strings["chat_playerJoin"] ) ~= nil then
		songbookWindow:PlayerJoined( sMessage );
		return;
	end
	if string.find( sMessage, Strings["chat_playerLeave"] ) ~= nil then
		songbookWindow:PlayerLeft( sMessage );
		return;
	end

	local temp, sPlayerName, sTrackName;
	temp, temp, sPlayerName, sTrackName = string.find( sMessage, Strings["chat_playReadyMsg"] );
	if not sPlayerName or not sTrackName then
		sPlayerName = songbookWindow.sPlayerName
		temp, temp, sTrackName = string.find( sMessage, Strings["chat_playSelfReadyMsg"] );		
	end

	if sPlayerName and sTrackName and songbookWindow.aPlayers then
		--code ajouté ici
		Turbine.PluginData.Save(Turbine.DataScope.Character, "FILETOPLAY", Strings["cmd_play"] .. " \"" .. SongDB.Songs[selectedSongIndex].Filepath .. selectedSong .. "\" " , 
			  function( result, message )
			if ( result ) then
				Turbine.Shell.WriteLine( "<rgb=#00FF00>Music to play have been written to file, Ready to Sync !</rgb>");
			else
				Turbine.Shell.WriteLine( "<rgb=#FF0000>Music to play failed to save : " .. message .. "</rgb>" );
			end
			end);
		--
		if sPlayerName == songbookWindow.sPlayerName and songbookWindow.sPlayerName then
			--songbookWindow:StopTimer( ) -- in case it is still counting ...
			sPlayerName = songbookWindow.sPlayerName;
			if songbookWindow.bCheckInstrument and sTrackName then songbookWindow:CheckInstrument( sTrackName ); end
		end
		if not songbookWindow.aPlayers[ sPlayerName ] then --- Player not yet registered 
			songbookWindow.nPlayers = songbookWindow.nPlayers + 1
			songbookWindow:AddPlayerToList( sPlayerName ); -- add to player listbox
			songbookWindow:UpdateMaxPartCount( )
		end
		songbookWindow.aPlayers[ sPlayerName ] = sTrackName; -- and to player array with the track name
		songbookWindow:SetTrackColours( selectedTrack );
		songbookWindow:SetPlayerColours( );
		songbookWindow:UpdateSetupColours( );
	end
end


-- Create the filter UI elements: Edit boxes for player count, transcriber, mood, genre
function SongbookWindow:CreateFilterUI( args )
	self.editPartcount = self:CreateFilterEdit( 0  );
	self.cbPartcount = self:CreateFilterCheckbox( 0, Strings["filterParts"] );
	self.cbPartcount.CheckedChanged =
		function( sender, args )
			--self:SetMaxPartCount( sender:IsChecked( ) )
			self:UpdateSongs( )
		end

	self.editComposer = self:CreateFilterEdit( 20 );
	self.cbComposer = self:CreateFilterCheckbox( 20, Strings["filterArtist"] );

	self.editGenre = self:CreateFilterEdit( 40 ); --,
	self.cbGenre = self:CreateFilterCheckbox( 40, Strings["filterGenre"] );
	
	self.editMood = self:CreateFilterEdit( 60 ); --,
	self.cbMood = self:CreateFilterCheckbox( 60, Strings["filterMood"] );

	self.editAuthor = self:CreateFilterEdit( 80 );
	self.cbAuthor = self:CreateFilterCheckbox( 80, Strings["filterAuthor"] );
	
	-- Separator filters - dir listbox
	self.separatorFilters = Turbine.UI.Control();
	self.separatorFilters:SetParent( self.listContainer );
	self.separatorFilters:SetZOrder( 300 );
	self.separatorFilters:SetBackColor( Turbine.UI.Color(1, 0.15, 0.15, 0.15) );
	self.separatorFilters:SetPosition( 156, self.dirlistBox:GetTop( ) );
	self.separatorFilters:SetSize( 10, self.dirlistBox:GetHeight( ) );
	self.separatorFilters:SetVisible( false );

	self:CreatePartyListbox( );
	self:CreateSetupsListbox( );
end


-- create a filter edit
function SongbookWindow:CreateFilterEdit( par_yPos, par_fn )
	local edit = Turbine.UI.Lotro.TextBox( );
	edit:SetParent( self.listContainer );
	edit:SetPosition( 0, par_yPos );
	edit:SetSize( 80, 20 );
	edit:SetFont( Turbine.UI.Lotro.Font.Verdana14 );
	edit:SetMultiline( false );	
	edit:SetVisible( false );
	--edit.FocusLost = par_fn;
	edit.KeyDown = function( sender, keyargs )
		if( keyargs.Action == 162 ) then
			--if( edit:HasFocus( ) ) then
				self:UpdateSongs( );
			--end
		end
	end
	return edit;
end


-- Create a filter checkbox
function SongbookWindow:CreateFilterCheckbox( par_yPos, par_sText )
	-- search button
	local cb = Turbine.UI.Lotro.CheckBox();
	cb:SetParent( self.listContainer );
	cb:SetPosition( 82, par_yPos );
	cb:SetSize( 80, 20 );
	cb:SetText( par_sText );
	cb.CheckedChanged =
		function( sender, args )
			self:UpdateSongs( );
		end
	cb:SetVisible( false );
	return cb;
end


-- Switch between filter UI display (true) and track listbox (false)
function SongbookWindow:ShowFilterUI( bFilter )
	self.bFilter = bFilter;
	self.editPartcount:SetVisible( bFilter ); self.cbPartcount:SetVisible( bFilter );
	self.editComposer:SetVisible( bFilter ); self.cbComposer:SetVisible( bFilter );
	self.editGenre:SetVisible( bFilter ); self.cbGenre:SetVisible( bFilter );
	self.editMood:SetVisible( bFilter ); self.cbMood:SetVisible( bFilter );
	self.editAuthor:SetVisible( bFilter ); self.cbAuthor:SetVisible( bFilter );
	self.separatorFilters:SetVisible( bFilter );
	
	self.btnParty:SetVisible( bFilter );
	self.listboxPlayers:SetVisible( bFilter );
	
	self:ResizeAll( );
end


-- Reposition the filter UI for dir listbox size changes
function SongbookWindow:AdjustFilterUI( )
	if not self.bFilter then return; end
	--if( not self.cbFilters:IsChecked( ) ) then return; end
	
	local dirHeight = self.dirlistBox:GetHeight( );
	if( dirHeight < 40 ) then dirHeight = 40; end

	self.editAuthor:SetVisible( dirHeight >= 100 ); self.cbAuthor:SetVisible( dirHeight >= 100 );
	self.editMood:SetVisible( dirHeight >= 80 ); self.cbMood:SetVisible( dirHeight >= 80 );
	self.editGenre:SetVisible( dirHeight >= 60 ); self.cbGenre:SetVisible( dirHeight >= 60 );

	self.separatorFilters:SetHeight( dirHeight );
end


function SongbookWindow:CheckInstrument( sTrack )
	local player = Turbine.Gameplay.LocalPlayer:GetInstance( );
	if not player then return; end
	local equip = player:GetEquipment( )
	if not equip then return; end
	local item = equip:GetItem( Turbine.Gameplay.Equipment.Instrument )
	if not item then return; end
	sTrack = sTrack:lower()
	--local sName = string.match( item:GetName( ), "%a+$" )
	self.bInstrumentOk = true -- only set to false if we can successfully determine track and equipped instrument
	local iName = self:GetInstrumentName( item:GetName( ):lower( ) )
	if not iName then return; end -- can't determine equipped instrument, disable message

	local iTrackInstrument = self:CheckTracksForInstrument( sTrack, iName, self.aInstruments ) -- try english names first
	if not iTrackInstrument then -- try localized names
		iTrackInstrument = self:CheckTracksForInstrument( sTrack, iName, aInstrumentsLoc )
	end

	if not iTrackInstrument then return; end -- could not determine the track instrument
	self:SetInstrumentMessage( aInstrumentsLoc[ iTrackInstrument ] ) -- print the localized name
end

function SongbookWindow:GetInstrumentName( sItem )
	if self.aSpecialInstruments and self.aSpecialInstruments[ sItem ] then
		return self.aSpecialInstruments[ sItem ] -- already contains the index 
	end
	for k,v in pairs( aInstrumentsLoc ) do
		if sItem:find( v ) then return k; end
	end
	return nil
end

function SongbookWindow:CheckTracksForInstrument( sTrack, iInstrument, aInstruments )
	if not iInstrument or iInstrument > #aInstruments then return nil; end
	local sName = aInstruments[ iInstrument ]
	for k,v in pairs( aInstruments ) do
		if sTrack:find( v ) then -- track name seems to contain the instrument name
			self.bInstrumentOk = not not string.find( sTrack, "[^%a]" .. sName:lower() )
			return k
		end
	end
	return nil
end
	

function SongbookWindow:SetInstrumentMessage( sInstr )
	if self.bInstrumentOk then
		self.tracksMsg:SetForeColor( self.colourDefault )
		self.tracksMsg:SetVisible( false )
	else
		self.tracksMsg:SetForeColor( self.colourWrongInstrument )
		self.tracksMsg:SetText( sInstr .. Strings["instr"] )
		self.tracksMsg:SetVisible( true )
	end
end


-- Add player to the fellowship list
function SongbookWindow:AddPlayerToList( sPlayername )
	if( self.listboxPlayers == nil ) then return; end -- Listbox not created yet

	local memberItem = Turbine.UI.Label( );
	memberItem:SetText( sPlayername );
	memberItem:SetTextAlignment( Turbine.UI.ContentAlignment.MiddleLeft );
	memberItem:SetSize( 100, 20 );				
	self.listboxPlayers:AddItem( memberItem );
end

-- Remove player from the fellowship list
function SongbookWindow:RemovePlayerFromList( sPlayername )
	if( self.listboxPlayers == nil ) then return; end -- Listbox not created yet

	local i, item = self:GetItemFromList( self.listboxPlayers, sPlayername )
	if i then self.listboxPlayers:RemoveItemAt( i ); end
end

-- Get index and item for player name from player listbox
function SongbookWindow:GetItemFromList( listbox, sText )
	local i, item
	for i = 1, listbox:GetItemCount( ) do
		item = listbox:GetItem( i )
		if item:GetText( ) == sText then return i, item; end
	end
	return nil, nil
end
		


function SongbookWindow:EnableReadyColumns( bOn )
	self.listboxPlayers:EnableCharColumn( bOn )
	self.tracklistBox:EnableCharColumn( bOn )
end

function SongbookWindow:ShowReadyColumns( bShow )
	if self.bShowReadyChars == bShow then return; end
	self.bShowReadyChars = bShow
	self:EnableReadyColumns( bShow )
	if selectedSongIndex then
		--self:ListTracks( selectedSongIndex )
		--self:RefreshPlayerListbox( )
		self:SetTrackColours( selectedTrack );
		self:SetPlayerColours( );
	end
end

function SongbookWindow:HightlightReadyColumns( bOn )
	self.listboxPlayers.bHighlightReadyCol = bOn
	self.tracklistBox.bHighlightReadyCol = bOn
end


-- Read party member names and add them to the party members listbox
-- TODO: Party object does not seem to report party members correctly?
function SongbookWindow:RefreshPlayerListbox( )
	if self.listboxPlayers == nil then return; end
	
	self.listboxPlayers:ClearItems( );

	local player = Turbine.Gameplay.LocalPlayer:GetInstance( );
	if player == nil then return; end
	
	if self.sPlayerName == nil then self.sPlayerName = player:GetName( ); end
	
	local party = player:GetParty( );
	if party == nil or party:GetMemberCount( ) <= 0 then
		if self.bChiefMode then	self.aPlayers = { }; self.aCurrentSongReady = { }; end
		self:AddPlayerToList( self.sPlayerName )
		self.aPlayers[ self.sPlayerName ] = 0
		return;
	end 

	-- If in chief mode, we rely on the party object; otherwise, we keep the known players.
	if self.bChiefMode then	self.aPlayers = { }; self.aCurrentSongReady = { }
	else self:ListKnownPlayers( ); end
	
	local iPlayer;
	for iPlayer = 1, party:GetMemberCount( ) do
		local member = party:GetMember( iPlayer )
		local sName = member:GetName( )
		if self.aPlayers[ sName ] == nil then self:AddPlayer( sName ); end
	end

	self:SetPlayerColours( ); -- restore current states
	self:UpdateMaxPartCount( )
	if self.maxPartCount then self:UpdateSongs( ) end
	self:UpdateSetupColours( );
end

-- Add player to arrays
function SongbookWindow:AddPlayer( sName )
	if self.aPlayers[ sName ] then return; end
	self.aCurrentSongReady[ sName ] = false
	self:AddPlayerToList( sName )
	self.aPlayers[ sName ] = 0
end

-- Add player to arrays
function SongbookWindow:RemovePlayer( sName )
	if not self.aPlayers[ sName ] then return; end
	self.aCurrentSongReady[ sName ] = nil
	self:RemovePlayerFromList( sName )
	self.aPlayers[ sName ] = nil
end


-- Write known players to player listbox
function SongbookWindow:ListKnownPlayers( )
	if not self.aPlayers then return; end
	for k,v in pairs( self.aPlayers ) do
		self:AddPlayerToList( k )
	end
end


-- Parse player join message, add player
-- Party object occasionally seems to become stale in raid settings, so we try to use client messages for player list updates
function SongbookWindow:PlayerJoined( sMsg )
	local temp, sPlayerName, sTrackName;
	temp, temp, sPlayerName = string.find( sMsg, "(%a+)" .. Strings["chat_playerJoin"] );
	if sPlayerName then self:AddPlayer( sPlayerName ); end
end

-- Parse player left message, remove player
function SongbookWindow:PlayerLeft( sMsg )
	local temp, sPlayerName, sTrackName;
	temp, temp, sPlayerName = string.find( sMsg, "(%a+)" .. Strings["chat_playerLeave"] );
	if sPlayerName then self:RemovePlayer( sPlayerName ); end
end


-- Clear the ready states for players
function SongbookWindow:ClearPlayerStates( )
	if not self.aPlayers then return; end
	for k in pairs( self.aPlayers ) do
		self.aPlayers[ k ] = 0; -- present, no song ready
	end
end


-- Update item colours in party listbox to indicate ready states
function SongbookWindow:SetPlayerColours( )
	if not self.aPlayers or not self.listboxPlayers then return; end

	local iMember;
	for iMember=1,self.listboxPlayers:GetItemCount( ) do
		local item = self.listboxPlayers:GetItem( iMember );
		if( self.aPlayers[ item:GetText( ) ] == nil ) then	-- should not happen
			item:SetForeColor( self.colourDefault );
			if self.bShowReadyChars then self.listboxPlayers:SetColumnChar( iMember, self.chNone, false ); end
		elseif( self.aPlayers[ item:GetText( ) ] == 0 ) then -- present, but no song ready
			item:SetForeColor( self.colourDefault );
			if self.bShowReadyChars then self.listboxPlayers:SetColumnChar( iMember, self.chNone, false ); end
		elseif self.aCurrentSongReady and self.aCurrentSongReady[ item:GetText( ) ] == 1 then 
			item:SetForeColor( self.colourReady ); -- Track from the currently displayed song ready
			if self.bShowReadyChars then self.listboxPlayers:SetColumnChar( iMember, self.chReady, false ); end
		elseif self.aCurrentSongReady and self.aCurrentSongReady[ item:GetText( ) ] == 2 then 
			item:SetForeColor( self.colourReadyMultiple ); -- Correct song, but same track as another player
			if self.bShowReadyChars then self.listboxPlayers:SetColumnChar( iMember, self.chMultiple, true ); end
		elseif self.aCurrentSongReady and self.aCurrentSongReady[ item:GetText( ) ] == 3 then 
			item:SetForeColor( self.colourDifferentSetup ); -- Correct song, but track not in current setup
			if self.bShowReadyChars then self.listboxPlayers:SetColumnChar( iMember, self.chWrongPart, true ); end
		else
			item:SetForeColor( self.colourDifferentSong ); -- Track ready, but different song
			if self.bShowReadyChars then self.listboxPlayers:SetColumnChar( iMember, self.chWrongSong, true ); end
		end
	end
end

-- Create party member listbox
function SongbookWindow:CreatePartyListbox( )
	local songsHeight = self.songlistBox:GetHeight( );
	local songsTop = self.songlistBox:GetTop( );
	if( songsHeight < 40 ) then songsHeight = 40; end

	-- Button to update party member list
	self.btnParty = Turbine.UI.Lotro.Button();
	self.btnParty:SetParent(self.listContainer);
	self.btnParty:SetPosition( 0, songsTop );
	self.btnParty:SetSize( 92, 20 );
	self.btnParty:SetText( Strings["players"] );
	self.btnParty:SetVisible( false );

	self.btnParty.MouseClick =
		function( sender, args )
			self:RefreshPlayerListbox( );
		end

	local dirsHeight = self.dirlistBox:GetHeight( );
	self.listboxPlayers = ListBoxCharColumn:New( 10, 20 );
	self.listboxPlayers:SetParent( self.listContainer );
	self.listboxPlayers:SetSize( 80, songsHeight - 20 );
	self.listboxPlayers:SetPosition( 2 , songsTop + 20 );
	self.listboxPlayers:SetVisible( false );
end

-- Use the number of currently listed players as limit for part counts
function SongbookWindow:SetMaxPartCount( bActivate )
	--self:RefreshPlayerListbox( )
	if bActivate and self.listboxPlayers then self.maxPartCount = self.listboxPlayers:GetItemCount( )
	else self.maxPartCount = nil;
	end
end

-- Due to party object issues, we only increase partcount here
function SongbookWindow:UpdateMaxPartCount( )
	if self.maxPartCount and self.listboxPlayers and self.maxPartCount < self.listboxPlayers:GetItemCount( ) then
		self.maxPartCount = self.listboxPlayers:GetItemCount( )
	end
end

-- Create listbox to show the possible setup counts
function SongbookWindow:CreateSetupsListbox( )
	self.listboxSetups = ListBoxScrolled:New( 10 );
	self.listboxSetups:SetParent( self.listContainer );
	self.listboxSetups:SetSize( self.listboxSetupsWidth - 0, self.tracklistBox:GetHeight( ) );
	self.listboxSetups:SetPosition( 0, self.tracklistBox:GetTop() );
	self.listboxSetups:SetVisible( self.bShowSetups );
	self.listboxSetups.SelectedIndexChanged =
		function( sender, args )
			self:ListTracksForSetup( sender:GetSelectedIndex( ) );
		end
end

function SongbookWindow:SelectSetup( iSetup )
	if not self.listboxSetups then return; end
	if not iSetup or iSetup > self.listboxSetups:GetItemCount( ) then iSetup = self.listboxSetups:GetItemCount( ); end
	self.listboxSetups:SetSelectedIndex( iSetup )
	self:ListTracksForSetup( iSetup )
end

function SongbookWindow:ListTracksForSetup( iSetup )
	if not SongDB.Songs[selectedSongIndex] or not SongDB.Songs[selectedSongIndex].Setups then return; end
	
	for iItem = 1, self.listboxSetups:GetItemCount( ) do
		self.listboxSetups:GetItem( iItem ):SetBackColor( self.backColourDefault );
	end

	local selTrack = self.tracklistBox:GetSelectedIndex( );
	
	self.aSetupTracksIndices = { };
	self.aSetupListIndices = { };
	self.iCurrentSetup = nil;

	if not iSetup or iSetup >= self.listboxSetups:GetItemCount( ) then
		self:ListTracks( selectedSongIndex );
		self.selectedSetupCount = nil
	else
		self.iCurrentSetup = iSetup;
		self.tracklistBox:ClearItems( );
		for i = 1, #SongDB.Songs[selectedSongIndex].Setups[ iSetup ] do
			local iTrack = SongDB.Songs[selectedSongIndex].Setups[ iSetup ]:byte( i ) - 64;
			self.aSetupTracksIndices[ i ] = iTrack;
			self.aSetupListIndices[ iTrack ] = i;
			self:AddTrackToList( selectedSongIndex, iTrack )
		end
		self.selectedSetupCount = #SongDB.Songs[selectedSongIndex].Setups[ iSetup ]
	end

	local selItem = self.listboxSetups:GetSelectedItem( );
	if selItem then selItem:SetBackColor( self.backColourHighlight ); end
	
	self:SelectTrack( 1 ); --selTrack );
	self:SetPlayerColours( );
	local found = self.tracklistBox:GetItemCount( );
	self.sepSongsTracks.heading:SetText( Strings["ui_parts"] .. " (" .. found .. ")" );
end

function SongbookWindow:SetupIndexForCount( iSong, setupCount )
	if not setupCount or not SongDB.Songs[iSong] or not SongDB.Songs[iSong].Setups then return nil; end
	for i = 1, #SongDB.Songs[iSong].Setups do
		if setupCount == #SongDB.Songs[iSong].Setups[ i ] then return i; end
	end
	return i
end

function SongbookWindow:UpdateSetupColours(  )
	if not self.listboxSetups or not SongDB.Songs[selectedSongIndex] or not SongDB.Songs[selectedSongIndex].Setups then return; end
	
	self:UpdateTrackReadyString( );

	local item;
	local matchPattern;
	local antiMatchPattern;
	local matchLength = 0;
	for i = 1, self.listboxSetups:GetItemCount( ) - 1 do
		item = self.listboxSetups:GetItem( i );

		matchPattern = "[" .. SongDB.Songs[selectedSongIndex].Setups[ i ] .. "]";
		antiMatchPattern = "[^" .. SongDB.Songs[selectedSongIndex].Setups[ i ] .. "]";
		_, matchLength = string.gsub( self.aReadyTracks, matchPattern, " " )

		if SongDB.Songs[selectedSongIndex].Setups[ i ] == self.aReadyTracks then
			item:SetForeColor( self.colourReady );
		elseif string.match( self.aReadyTracks, antiMatchPattern ) then
			item:SetForeColor( Turbine.UI.Color( 0.7, 0, 0 ) );
		elseif matchLength and matchLength + 1 == #SongDB.Songs[selectedSongIndex].Setups[ i ] then
			item:SetForeColor( Turbine.UI.Color( 0, 0.7, 0 ) );
		else
			item:SetForeColor( self.colourDefault );
		end
	end
end

function SongbookWindow:ClearSetups(  )
	if not self.listboxSetups then return; end	
	local selItem = self:SetListboxColours( self.listboxSetups, true );
	if selItem then selItem:SetBackColor( self.backColourHighlight ); end
end

function SongbookWindow:UpdateTrackReadyString( )
	self.aReadyTracks = "";
	for iList = 1,self.tracklistBox:GetItemCount( ) do
		local i = self:SelectedTrackIndex( iList );
		if self:GetTrackReadyState( SongDB.Songs[selectedSongIndex].Tracks[i].Name ) then
			self.aReadyTracks = self.aReadyTracks .. string.char( 0x40 + i );
		end
	end
end
		

-- Add an item with the given text to the given listbox
function SongbookWindow:AddItemToList( sText, listbox, width )
	if( listbox == nil ) then return; end -- Listbox not created yet

	local item = Turbine.UI.Label( );
	item:SetText( sText );
	item:SetTextAlignment( Turbine.UI.ContentAlignment.MiddleCenter );
	item:SetSize( width, 20 );				
	listbox:AddItem(item);
end


function SongbookWindow:ListSetups( songID )
	if not self.listboxSetups or not self.listboxPlayers then return; end
	self.listboxSetups:ClearItems( )
	if not SongDB or not SongDB.Songs[songID] or not SongDB.Songs[songID].Setups then
		self:ShowSetups( false )
		return;
	end

	self:ShowSetups( true )
	local playerCount;
	if not self.maxPartCount then playerCount = 1000;
	else playerCount = self.maxPartCount; end
--playerCount = 6;	
	local countInSetup;
	for i = 1, #SongDB.Songs[songID].Setups do
		countInSetup = #SongDB.Songs[songID].Setups[ i ];
		if countInSetup <= playerCount then
			self:AddItemToList( countInSetup, self.listboxSetups, self.listboxSetupsWidth );
		end
	end
	self:AddItemToList( "A", self.listboxSetups, self.listboxSetupsWidth );
end

-- Return a track state indicator: 
-- nil = track not ready, name of a player = ready by this player, 0 = ready by more than one player
function SongbookWindow:GetTrackReadyState( sTrackName, indicator )
	local found = nil;
	local readyIndicator = 1;
	if indicator then readyIndicator = indicator; end
	if( self.aPlayers ~= nil ) then
		for k,v in pairs( self.aPlayers ) do
			if( v == string.sub( sTrackName, 1, 63 ) ) then
				if found == nil then
					found = k;
					self.aCurrentSongReady[ k ] = readyIndicator; -- track ready
				else
					self.aCurrentSongReady[ k ] = 2; -- track ready, but by other player too
					if found ~= 0 then
						self.aCurrentSongReady[ found ] = 2; -- set for other player as well
						found = 0;
					end
				end
			end
		end
	end
	
	return found; -- nil if not ready, player name if ready once, 0 if ready more than once
end


function SongbookWindow:ClearPlayerReadyStates( )
	if( self.aPlayers ~= nil ) then
		for k,v in pairs( self.aPlayers ) do
			self.aCurrentSongReady[ k ] = false;
		end
	end
end


function SongbookWindow:SetChiefMode( bState )
	self.bChiefMode = (bState==true)
	self.syncStartSlot:SetVisible( self.bChiefMode )
	self.syncStartIcon:SetVisible( self.bChiefMode )
end


-- ListBoxScrolled ----------------------------------------
-- a Listbox with child scrollbar and separator

function ListBoxScrolled:New( scrollWidth, listbox )
	listbox = listbox or ListBoxScrolled( scrollWidth );
	setmetatable( listbox, self );
	self.__index = self;
	return listbox;
end

function ListBoxScrolled:Constructor( scrollWidth )
	Turbine.UI.ListBox.Constructor( self );
	self:SetMouseVisible( true );	
	self.scrollWidth = scrollWidth;
	self:CreateChildScrollbar( scrollWidth );
	self:CreateChildSeparator( scrollWidth );
end

function ListBoxScrolled:CreateChildScrollbar( width )
	self.scrollBar = Turbine.UI.Lotro.ScrollBar( );
	self.scrollBar:SetParent( self:GetParent( ) );
	self.scrollBar:SetOrientation( Turbine.UI.Orientation.Vertical );
	self.scrollBar:SetZOrder( 320 );
	self.scrollBar:SetWidth( width );
	self.scrollBar:SetTop( 0 );
	self.scrollBar:SetValue( 0 );
	self:SetVerticalScrollBar( self.scrollBar );
	self.scrollBar:SetVisible( false );
end

function ListBoxScrolled:CreateChildSeparator( width )
	self.separator = Turbine.UI.Control( );
	self.separator:SetParent( self:GetParent( ) );
	self.separator:SetZOrder( 300 );
	self.separator:SetWidth( width );
	self.separator:SetTop( 0 );
	self.separator:SetBackColor(Turbine.UI.Color(1, 0.15, 0.15, 0.15));
	self.separator:SetVisible( false );
end
	
function ListBoxScrolled:SetLeft( xPos )
	Turbine.UI.ListBox.SetLeft( self, xPos );
	self.scrollBar:SetLeft( xPos + self:GetWidth( ) );
	self.separator:SetLeft( xPos + self:GetWidth( ) );
end
	
function ListBoxScrolled:SetTop( yPos )
	Turbine.UI.ListBox.SetTop( self, yPos );
	self.scrollBar:SetTop( yPos );
	self.separator:SetTop( yPos );
end
	
function ListBoxScrolled:SetPosition( xPos, yPos )
	self:SetLeft( xPos );
	self:SetTop( yPos );
end
	
function ListBoxScrolled:SetWidth( width )
	Turbine.UI.ListBox.SetWidth( self, width );
	self.scrollBar:SetLeft( self:GetLeft( ) + width );
	self.separator:SetLeft( self:GetLeft( ) + width );
end

function ListBoxScrolled:SetHeight( height )
	Turbine.UI.ListBox.SetHeight( self, height );
	self.scrollBar:SetHeight( height );
	self.separator:SetHeight( height );
end

function ListBoxScrolled:SetSize( width, height )
	self:SetWidth( width );
	self:SetHeight( height );
end

function ListBoxScrolled:SetVisible( bVisible )
	Turbine.UI.ListBox.SetVisible( self, bVisible );
	self.separator:SetVisible( bVisible );
	self.scrollBar:SetVisible( bVisible );
	if bVisible then self.scrollBar:SetParent( self:GetParent( ) )
	else self.scrollBar:SetParent( self ); end
end

function ListBoxScrolled:SetParent( parent )
	Turbine.UI.ListBox.SetParent( self, parent );
	self.separator:SetParent( parent );
	self.scrollBar:SetParent( parent );
end

-- /ListBoxScrolled ----------------------------------------


-- ListBoxReadyIndicator ----------------------------------------
-- A scroll listbox with an optional single-char column before the main column

function ListBoxCharColumn:New( scrollWidth, readyColWidth, listbox )
	listbox = listbox or ListBoxCharColumn( scrollWidth, readyColWidth );
	setmetatable( listbox, self );
	self.__index = self;
	return listbox;
end

function ListBoxCharColumn:Constructor( scrollWidth, readyColWidth )
	ListBoxScrolled.Constructor( self, scrollWidth );
	self.readyColWidth = readyColWidth
	self.bShowReadyChars = false
	self.bHighlightReadyCol = false
end

function ListBoxCharColumn:EnableCharColumn( bColumn )
	if self.bShowReadyChars == bColumn then return; end

	self.bShowReadyChars = bColumn
	if ListBoxScrolled.GetItemCount( self ) == 0 then return; end

	local iList
	local itemCount = ListBoxScrolled.GetItemCount( self )
	if bColumn then -- Add a char item before every item in the list
		for iList = 1, itemCount * 2, 2 do
			local chItem = self:CreateCharItem( )
			ListBoxScrolled.InsertItem( self, iList, chItem )
		end
		self:SetMaxItemsPerLine( 2 )
	else -- remove every item with odd index (the char items)
		for iList = 1, itemCount / 2 do
			ListBoxScrolled.RemoveItemAt( self, iList )
		end
		self:SetMaxItemsPerLine( 1 )
	end
end

function ListBoxCharColumn:GetItem( iLine )
	if self.bShowReadyChars then iLine = iLine * 2; end
	return ListBoxScrolled.GetItem( self, iLine )
end

function ListBoxCharColumn:GetCharColumnItem( iLine )
	if not self.bShowReadyChars then return nil; end
	return ListBoxScrolled.GetItem( self, iLine * 2 - 1 )
end

function ListBoxCharColumn:SetColumnChar( iLine, ch, bHighlight )
	local charItem = self:GetCharColumnItem( iLine )
	if charItem then
		self:ApplyHighlight( charItem, bHighlight )
		charItem:SetText( ch )
	end
end

function ListBoxCharColumn:ApplyHighlight( charItem, bHighlight )
	if bHighlight and self.bHighlightReadyCol then
		charItem:SetForeColor( Turbine.UI.Color( 1, 0, 0, 0 ) )
		charItem:SetBackColor( Turbine.UI.Color( 1, 0.7, 0.7, 0.7 ) )
	else
		charItem:SetForeColor( Turbine.UI.Color( 1, 1, 1, 1 ) )
		charItem:SetBackColor( Turbine.UI.Color( 1, 0, 0, 0 ) )
	end
end

function ListBoxCharColumn:GetItemCount( )
	if self.bShowReadyChars then return math.floor( ListBoxScrolled.GetItemCount( self ) / 2 ); end
	return ListBoxScrolled.GetItemCount( self )
end

function ListBoxCharColumn:ClearItems( )
	ListBoxScrolled.ClearItems( self )
	if self.bShowReadyChars then self:SetMaxItemsPerLine( 2 )
	else self:SetMaxItemsPerLine( 1 ); end
	self:SetOrientation( Turbine.UI.Orientation.Horizontal )
end

function ListBoxCharColumn:CreateCharItem( )
	local charItem = Turbine.UI.Label( )
	charItem:SetMultiline( false )
	charItem:SetSize( self.readyColWidth, 20 )
	charItem:SetTextAlignment( Turbine.UI.ContentAlignment.MiddleCenter )
	self:ApplyHighlight( charItem, false )
	return charItem
end

function ListBoxCharColumn:AddItem( item )
	if self.bShowReadyChars then -- add ready indicator (single char in first column)
		local charItem = self:CreateCharItem( )
		ListBoxScrolled.AddItem( self, charItem )			
	end
	ListBoxScrolled.AddItem( self, item )
end

function ListBoxCharColumn:RemoveItemAt( i )
	if self.bShowReadyChars then
		ListBoxScrolled.RemoveItemAt( self, i * 2 )
		ListBoxScrolled.RemoveItemAt( self, i * 2 - 1 )
	else
		ListBoxScrolled.AddItem( self, i )
	end
end

-- /ListBoxReadyIndicator ----------------------------------------
	
	
function SongbookWindow:SetListboxColours( listbox, bNoSelectionHighlight )
	for i = 1,listbox:GetItemCount() do
		local item = listbox:GetItem( i );	
		item:SetForeColor( self.colourDefault );
	end
	if bNoSelectionHighlight then return nil; end 
	local selectedItem = listbox:GetSelectedItem( )
	if selectedItem then selectedItem:SetForeColor( self.colourDefaultHighlighted ); end
	return selectedItem;
end

function SongbookWindow:CreateSeparator(left, top, width, height)
	local separator = Turbine.UI.Control();
	separator:SetParent(self.listContainer);
	separator:SetZOrder(300);
	separator:SetBackColor(Turbine.UI.Color(1, 0.15, 0.15, 0.15));
	separator:SetPosition( left, top );
	separator:SetSize(width,height);
	separator:SetVisible( false );
	return separator;
end

function SongbookWindow:CreateMainSeparator(top)
	return self:CreateSeparator(0, top, self.listContainer:GetWidth(), 13);
end

function SongbookWindow:CreateSeparatorHeading(parent, sText)
	local heading = Turbine.UI.Label();
	heading:SetParent(parent);
	heading:SetLeft(0);
	heading:SetSize(100,13);
	heading:SetFont(Turbine.UI.Lotro.Font.TrajanPro13);
	heading:SetText(sText);
	heading:SetMouseVisible(false);
	return heading;
end

function SongbookWindow:CreateSeparatorArrows(parent)
	local arrows = Turbine.UI.Control();
	arrows:SetParent(parent);
	arrows:SetZOrder(310);
	arrows:SetBackground(gDir .. "arrows.tga");
	arrows:SetSize(20,10);
	arrows:SetPosition(parent:GetWidth( ) / 2 - 10, 1);
	arrows:SetMouseVisible( false );
	return arrows;
end

function SongbookWindow:CreateMainShortcut(left)
	local slot = Turbine.UI.Lotro.Quickslot();
	slot:SetParent(self);
	slot:SetPosition(left, 50);
	slot:SetSize(32, 30);
	slot:SetZOrder(100);
	slot:SetAllowDrop(false);
	slot:SetVisible( true );
	return slot;
end

function SongbookWindow:CreateMainIcon(left,sImageName)
	local icon = Turbine.UI.Control();
	icon:SetParent( self );
	icon:SetPosition(left, 50);
	icon:SetSize(35, 35);
	icon:SetZOrder(110);
	icon:SetMouseVisible(false);
	icon:SetBlendMode( Turbine.UI.BlendMode.AlphaBlend );
	icon:SetBackground(gDir .. sImageName .. ".tga" );
	return icon;
end

function SongbookWindow:AddScrollbar(parent,listbox,xPos,yPos)
	local scroll = Turbine.UI.Lotro.ScrollBar();
	scroll:SetParent( parent );
	scroll:SetOrientation( Turbine.UI.Orientation.Vertical );
	scroll:SetPosition( xPos, yPos )
	scroll:SetHeight( listbox:GetHeight() );
	scroll:SetZOrder(320);
	scroll:SetWidth(10);
	scroll:SetValue(0);
	listbox:SetVerticalScrollBar( scroll );
	scroll:SetVisible( false );
	return scroll;
end


function SongbookWindow:SetSonglistHeight( height )
	self.songlistBox:SetHeight( height );
	self.sepSongsTracks:SetTop(self.listContainer:GetHeight() - self.tracklistBox:GetHeight() - 13);
	self.listboxPlayers:SetHeight( height - 20 );
end

function SongbookWindow:AdjustSonglistHeight( )
	local height = self.listContainer:GetHeight() - self.dirlistBox:GetHeight() - 13;
	if Settings.TracksVisible == "yes" then height = height - self.tracklistBox:GetHeight() - 13; end
	self:SetSonglistHeight( height )
end

function SongbookWindow:SetSonglistTop( top )
	self.songlistBox:SetTop( top );
	self.separator1:SetTop( top - 13 );
	self.btnParty:SetTop( top );
	self.listboxPlayers:SetTop( top + 20 );
end

function SongbookWindow:AdjustSonglistLeft( )
	local xPos = 0;
	if self.bFilter and self.bShowPlayers then xPos = xPos + 95; end
	self.songlistBox:SetLeft( xPos );
	self.songlistBox:SetWidth( self.listContainer:GetWidth( ) - xPos - 10 );
end

function SongbookWindow:AdjustSonglistPosition( )
	self:AdjustSonglistLeft( )
	self:SetSonglistTop( self.dirlistBox:GetHeight() + 13 )
	self.separator1:SetWidth( self.listContainer:GetWidth() );
	self.sArrows1:SetLeft( self.separator1:GetWidth() / 2 - 10 );
end

function SongbookWindow:AdjustDirlistSize( )
	local width = self.listContainer:GetWidth( ) - 10;
	local height = self.listContainer:GetHeight() - self.songlistBox:GetHeight() - 13;

	if self.bFilter then width = width - 170; end
	if Settings.TracksVisible == "yes" then height = height - Settings.TracksHeight - 13; end

	self.dirlistBox:SetSize( width, height );
end

function SongbookWindow:AdjustDirlistPosition( )
	local xPos = 0;
	if self.bFilter then xPos = xPos + 170; end
	self.dirlistBox:SetPosition( xPos , 0 );
	self.dirlistBox:SetWidth( self.listContainer:GetWidth( ) - xPos - 10 );
end

function SongbookWindow:ShowTrackListbox( bShow )
	self.tracklistBox:SetVisible( bShow );
	self.sepSongsTracks:SetVisible( bShow );
	self.sArrows2:SetVisible( bShow );		
end

function SongbookWindow:SetTracklistTop( top )
	self.tracklistBox:SetTop( top )
	self.sepSongsTracks:SetTop( top - 13 )
	self.listboxSetups:SetTop( top );
	self.tracksMsg:SetTop( top - 15 );
end

function SongbookWindow:MoveTracklistTop( delta )
	self:SetTracklistTop( self.tracklistBox:GetTop( ) + delta )
end

function SongbookWindow:AdjustTracklistLeft( )
	if self.bShowSetups then self.tracklistBox:SetLeft( self.setupsWidth )
	else self.tracklistBox:SetLeft( 0 ) end
end

function SongbookWindow:AdjustTracklistWidth( )
	local width = self.listContainer:GetWidth( ) - 10
	if self.bShowSetups then width = width - self.setupsWidth end
	self.tracklistBox:SetWidth( width )
	--if self.alignTracksRight == true then self:AdjustTracklistItemsPosition( width ) end
	if self.alignTracksRight == true then self:RealignTracknames( ) end
	self.sepSongsTracks:SetWidth( self.listContainer:GetWidth( ) )
	self.sArrows2:SetLeft( self.sepSongsTracks:GetWidth() / 2 - 10 )
	self.tracksMsg:SetLeft( self.tracklistBox:GetLeft( ) + width - self.tracksMsg:GetWidth( ) )
end

function SongbookWindow:AdjustTracklistItemsPosition( width )
	for i = 1, self.tracklistBox:GetItemCount( ) do
		local item = self.tracklistBox:GetItem( i )
		item:SetLeft( width - 1000 )
	end
end

function SongbookWindow:SetTracklistHeight( height )
	self.tracklistBox:SetHeight( height )			
	self.listboxSetups:SetHeight( height );
end

function SongbookWindow:AdjustTracklistSize( height )
	self:AdjustTracklistWidth( )
	self:SetTracklistHeight( height )
end

function SongbookWindow:UpdateTracklistTop( )
	self:SetTracklistTop( self.listContainer:GetHeight() - self.tracklistBox:GetHeight() )
end

function SongbookWindow:ShowSetups( bShow )
	if bShow and not self.bShowSetups then 
		self.bShowSetups = true
		self.listboxSetups:SetVisible( true )
		self:AdjustTracklistLeft( )
		self:AdjustTracklistWidth( )
	elseif not bShow and self.bShowSetups then 
		self.bShowSetups = false
		self.listboxSetups:ClearItems( )
		self.listboxSetups:SetVisible( false )
		self:AdjustTracklistLeft( )
		self:AdjustTracklistWidth( )
	end
end


function AddCallback(object, event, callback)
    if (object[event] == nil) then
        object[event] = callback;
    else
        if (type(object[event]) == "table") then
            table.insert(object[event], callback);
        else
            object[event] = {object[event], callback};
        end
    end
    return callback;
end

function RemoveCallback(object, event, callback)
    if (object[event] == callback) then
        object[event] = nil;
    else
        if (type(object[event]) == "table") then
            local size = table.getn(object[event]);
            for i = 1, size do
                if (object[event][i] == callback) then
                    table.remove(object[event], i);
                    break;
                end
            end
        end
    end
end

--Turbine.Shell.WriteLine( "Checking track: " .. string.sub( sTrackName, 1, 63 ) );
